# Copyright 2019, 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose, Aniek Roelofs, Tuan Chien


from __future__ import annotations

import base64
import binascii
import json
import os
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Callable, ClassVar, Dict, List, TextIO, Tuple, Union

import cerberus.validator
import yaml
from cerberus import Validator
from cryptography.fernet import Fernet

from observatory.platform.cli.cli_utils import (
    INDENT1,
    INDENT3,
    comment,
    indent,
)
from observatory.platform.terraform.terraform_api import TerraformVariable
from observatory.platform.utils.airflow_utils import AirflowVars
from observatory.platform.utils.config_utils import (
    observatory_home as default_observatory_home,
)


def generate_fernet_key() -> str:
    """Generate a Fernet key.

    :return: A newly generated Fernet key.
    """

    return Fernet.generate_key().decode("utf8")


def generate_secret_key(length: int = 30) -> str:
    """Generate a secret key for the Flask Airflow Webserver.

    :param length: the length of the key to generate.
    :return: the random key.
    """

    return binascii.b2a_hex(os.urandom(length)).decode("utf-8")


def save_yaml(file_path: str, dict_: Dict):
    """Save a yaml file from a dictionary.

    :param file_path: the path to the file to save.
    :param dict_: the dictionary.
    :return: None.
    """

    with open(file_path, "w") as yaml_file:
        yaml.dump(dict_, yaml_file, default_flow_style=False)


def to_hcl(value: Dict) -> str:
    """Convert a Python dictionary into HCL.

    :param value: the dictionary.
    :return: the HCL string.
    """

    return json.dumps(value, separators=(",", "="))


def from_hcl(string: str) -> Dict:
    """Convert an HCL string into a Dict.

    :param string: the HCL string.
    :return: the dict.
    """

    return json.loads(string.replace('"=', '":'))


class BackendType(Enum):
    """The type of backend"""

    local = "local"
    terraform = "terraform"


class Environment(Enum):
    """The environment being used"""

    develop = "develop"
    staging = "staging"
    production = "production"


@dataclass
class Backend:
    """The backend settings for the Observatory Platform.

    Attributes:
        type: the type of backend being used (local environment or Terraform).
        environment: what type of environment is being deployed (develop, staging or production).
    """

    type: BackendType = BackendType.local
    environment: Environment = Environment.develop

    @staticmethod
    def from_dict(dict_: Dict) -> Backend:
        """Constructs a Backend instance from a dictionary.

        :param dict_: the dictionary.
        :return: the Backend instance.
        """

        backend_type = BackendType(dict_.get("type"))
        environment = Environment(dict_.get("environment"))

        return Backend(
            backend_type,
            environment,
        )


@dataclass
class PythonPackage:
    name: str
    host_package: str
    docker_package: str
    type: str


@dataclass
class Observatory:
    """The Observatory settings for the Observatory Platform.

    Attributes:
        :param package: the observatory platform package, either a local path to a Python source package (editable type),
        path to a sdist (sdist) or a PyPI package name and version (pypi).
        :param package_type: the package type, editable, sdist, pypi.
        :param airflow_fernet_key: the Fernet key.
        :param airflow_secret_key: the secret key used to run the flask app.
        :param airflow_ui_user_password: the password for the Apache Airflow UI admin user.
        :param airflow_ui_user_email: the email address for the Apache Airflow UI admin user.
        :param observatory_home: The observatory home folder.
        :param postgres_password: the Postgres SQL password.
        :param redis_port: The host Redis port number.
        :param flower_ui_port: The host's Flower UI port number.
        :param airflow_ui_port: The host's Apache Airflow UI port number.
        :param elastic_port: The host's Elasticsearch port number.
        :param kibana_port: The host's Kibana port number.
        :param docker_network_name: The Docker Network name, used to specify a custom Docker Network.
        :param docker_network_is_external: whether the docker network is external or not.
        :param docker_compose_project_name: The namespace for the Docker Compose containers: https://docs.docker.com/compose/reference/envvars/#compose_project_name.
        :param enable_elk: whether to enable the elk stack or not.
        :param api_port: Port for accessing the API locally (exposed by docker container).
    """

    package: str = "observatory-platform"
    package_type: str = "pypi"
    airflow_fernet_key: str = field(default_factory=generate_fernet_key)
    airflow_secret_key: str = field(default_factory=generate_secret_key)
    airflow_ui_user_email: str = "airflow@airflow.com"
    airflow_ui_user_password: str = "airflow"
    observatory_home: str = default_observatory_home()
    postgres_password: str = "postgres"
    redis_port: int = 6379
    flower_ui_port: int = 5555
    airflow_ui_port: int = 8080
    elastic_port: int = 9200
    kibana_port: int = 5601
    docker_network_name: str = "observatory-network"
    docker_network_is_external: bool = False
    docker_compose_project_name: str = "observatory"
    enable_elk: bool = True
    api_package: str = "observatory-api"
    api_package_type: str = "pypi"
    api_port: int = 5002

    def to_hcl(self):
        return to_hcl(
            {
                "airflow_fernet_key": self.airflow_fernet_key,
                "airflow_secret_key": self.airflow_secret_key,
                "airflow_ui_user_password": self.airflow_ui_user_password,
                "airflow_ui_user_email": self.airflow_ui_user_email,
                "postgres_password": self.postgres_password,
            }
        )

    @property
    def host_package(self):
        return os.path.normpath(self.package)

    @staticmethod
    def from_dict(dict_: Dict) -> Observatory:
        """Constructs an Airflow instance from a dictionary.

        :param dict_: the dictionary.
        :return: the Airflow instance.
        """

        package = dict_.get("package")
        package_type = dict_.get("package_type")
        airflow_fernet_key = dict_.get("airflow_fernet_key")
        airflow_secret_key = dict_.get("airflow_secret_key")
        airflow_ui_user_email = dict_.get("airflow_ui_user_email", Observatory.airflow_ui_user_email)
        airflow_ui_user_password = dict_.get("airflow_ui_user_password", Observatory.airflow_ui_user_password)
        observatory_home = dict_.get("observatory_home", Observatory.observatory_home)
        postgres_password = dict_.get("postgres_password", Observatory.postgres_password)
        redis_port = dict_.get("redis_port", Observatory.redis_port)
        flower_ui_port = dict_.get("flower_ui_port", Observatory.flower_ui_port)
        airflow_ui_port = dict_.get("airflow_ui_port", Observatory.airflow_ui_port)
        elastic_port = dict_.get("elastic_port", Observatory.elastic_port)
        kibana_port = dict_.get("kibana_port", Observatory.kibana_port)
        docker_network_name = dict_.get("docker_network_name", Observatory.docker_network_name)
        docker_network_is_external = dict_.get("docker_network_is_external", Observatory.docker_network_is_external)
        docker_compose_project_name = dict_.get("docker_compose_project_name", Observatory.docker_compose_project_name)
        enable_elk = dict_.get("enable_elk", Observatory.enable_elk)
        api_package = dict_.get("api_package", Observatory.api_package)
        api_package_type = dict_.get("api_package_type", Observatory.api_package_type)
        api_port = dict_.get("api_port", Observatory.api_port)

        return Observatory(
            package,
            package_type,
            airflow_fernet_key,
            airflow_secret_key,
            airflow_ui_user_password=airflow_ui_user_password,
            airflow_ui_user_email=airflow_ui_user_email,
            observatory_home=observatory_home,
            postgres_password=postgres_password,
            redis_port=redis_port,
            flower_ui_port=flower_ui_port,
            airflow_ui_port=airflow_ui_port,
            elastic_port=elastic_port,
            kibana_port=kibana_port,
            docker_network_name=docker_network_name,
            docker_network_is_external=docker_network_is_external,
            docker_compose_project_name=docker_compose_project_name,
            enable_elk=enable_elk,
            api_package=api_package,
            api_package_type=api_package_type,
            api_port=api_port,
        )


@dataclass
class CloudStorageBucket:
    """Represents a Google Cloud Storage Bucket.

    Attributes:
        id: the id of the bucket (which gets set as an Airflow variable).
        name: the name of the Google Cloud storage bucket.
    """

    id: str
    name: str

    @staticmethod
    def parse_buckets(buckets: Dict) -> List[CloudStorageBucket]:
        return parse_dict_to_list(buckets, CloudStorageBucket)


@dataclass
class GoogleCloud:
    """The Google Cloud settings for the Observatory Platform.

    Attributes:
        project_id: the Google Cloud project id.
        credentials: the path to the Google Cloud credentials.
        region: the Google Cloud region.
        zone: the Google Cloud zone.
        data_location: the data location for storing buckets.
        buckets: a list of the Google Cloud buckets.
    """

    project_id: str = None
    credentials: str = None
    region: str = None
    zone: str = None
    data_location: str = None
    buckets: List[CloudStorageBucket] = field(default_factory=lambda: [])

    def read_credentials(self) -> str:
        with open(self.credentials, "r") as f:
            data = f.read()
        return data

    def to_hcl(self):
        return to_hcl(
            {
                "project_id": self.project_id,
                "credentials": self.read_credentials(),
                "region": self.region,
                "zone": self.zone,
                "data_location": self.data_location,
                "buckets": [bucket.name for bucket in self.buckets],
            }
        )

    @staticmethod
    def from_dict(dict_: Dict) -> GoogleCloud:
        """Constructs a GoogleCloud instance from a dictionary.

        :param dict_: the dictionary.
        :return: the GoogleCloud instance.
        """

        project_id = dict_.get("project_id")
        credentials = dict_.get("credentials")
        region = dict_.get("region")
        zone = dict_.get("zone")
        data_location = dict_.get("data_location")
        buckets = CloudStorageBucket.parse_buckets(dict_.get("buckets", dict()))

        return GoogleCloud(
            project_id=project_id,
            credentials=credentials,
            region=region,
            zone=zone,
            data_location=data_location,
            buckets=buckets,
        )


def parse_dict_to_list(dict_: Dict, cls: ClassVar) -> List[Any]:
    """Parse the key, value pairs in a dictionary into a list of class instances.

    :param dict_: the dictionary.
    :param cls: the type of class to construct.
    :return: a list of class instances.
    """

    parsed_items = []
    for key, val in dict_.items():
        parsed_items.append(cls(key, val))
    return parsed_items


@dataclass
class WorkflowsProject:
    """Represents a project that contains DAGs to load.

    Attributes:
        :param package_name: the package name.
        :param package: the observatory platform package, either a local path to a Python source package (editable type),
        path to a sdist (sdist) or a PyPI package name and version (pypi).
        :param package_type: the package type, editable, sdist, pypi.
        dags_module: the Python import path to the module that contains the Apache Airflow DAGs to load.
    """

    package_name: str
    package: str
    package_type: str
    dags_module: str

    @staticmethod
    def parse_workflows_projects(list: List) -> List[WorkflowsProject]:
        """Parse the workflows_projects list object into a list of WorkflowsProject instances.

        :param list: the list.
        :return: a list of WorkflowsProject instances.
        """

        parsed_items = []
        for item in list:
            package_name = item["package_name"]
            package = item["package"]
            package_type = item["package_type"]
            dags_module = item["dags_module"]
            parsed_items.append(WorkflowsProject(package_name, package, package_type, dags_module))
        return parsed_items


def list_to_hcl(items: List[Union[AirflowConnection, AirflowVariable]]) -> str:
    """Convert a list of AirflowConnection or AirflowVariable instances into HCL.

    :param items: the list of AirflowConnection or AirflowVariable instances.
    :return: the HCL string.
    """

    dict_ = dict()
    for item in items:
        dict_[item.name] = item.value
    return to_hcl(dict_)


@dataclass
class AirflowConnection:
    """An Airflow Connection.

    Attributes:
        name: the name of the Airflow Connection.
        value: the value of the Airflow Connection.
    """

    name: str
    value: str

    @property
    def conn_name(self) -> str:
        """The Airflow Connection environment variable name, which is required to set the connection from an
        environment variable.

        :return: the Airflow Connection environment variable name.
        """

        return f"AIRFLOW_CONN_{self.name.upper()}"

    @staticmethod
    def parse_airflow_connections(dict_: Dict) -> List[AirflowConnection]:
        """Parse the airflow_connections dictionary object into a list of AirflowConnection instances.

        :param dict_: the dictionary.
        :return: a list of AirflowConnection instances.
        """

        return parse_dict_to_list(dict_, AirflowConnection)


@dataclass
class AirflowVariable:
    """An Airflow Variable.

    Attributes:
        name: the name of the Airflow Variable.
        value: the value of the Airflow Variable.
    """

    name: str
    value: str

    @property
    def env_var_name(self):
        """The Airflow Variable environment variable name, which is required to set the variable from an
        environment variable.

        :return: the Airflow Variable environment variable name.
        """

        return f"AIRFLOW_VAR_{self.name.upper()}"

    @staticmethod
    def parse_airflow_variables(dict_: Dict) -> List[AirflowVariable]:
        """Parse the airflow_variables dictionary object into a list of AirflowVariable instances.

        :param dict_: the dictionary.
        :return: a list of AirflowVariable instances.
        """

        return parse_dict_to_list(dict_, AirflowVariable)


@dataclass
class Terraform:
    """The Terraform settings for the Observatory Platform.

    Attributes:
        organization: the Terraform Organisation name.
    """

    organization: str

    @staticmethod
    def from_dict(dict_: Dict) -> Terraform:
        """Constructs a Terraform instance from a dictionary.

        :param dict_: the dictionary.
        """

        organization = dict_.get("organization")
        return Terraform(organization)


@dataclass
class CloudSqlDatabase:
    """The Google Cloud SQL database settings for the Observatory Platform.

    Attributes:
        tier: the database machine tier.
        backup_start_time: the start time for backups in HH:MM format.
    """

    tier: str
    backup_start_time: str

    def to_hcl(self):
        return to_hcl({"tier": self.tier, "backup_start_time": self.backup_start_time})

    @staticmethod
    def from_dict(dict_: Dict) -> CloudSqlDatabase:
        """Constructs a CloudSqlDatabase instance from a dictionary.

        :param dict_: the dictionary.
        :return: the CloudSqlDatabase instance.
        """

        tier = dict_.get("tier")
        backup_start_time = dict_.get("backup_start_time")
        return CloudSqlDatabase(tier, backup_start_time)


@dataclass
class VirtualMachine:
    """A Google Cloud virtual machine.

    Attributes:
        machine_type: the type of Google Cloud virtual machine.
        disk_size: the size of the disk in GB.
        disk_type: the disk type; pd-standard or pd-ssd.
        create: whether to create the VM or not.
    """

    machine_type: str
    disk_size: int
    disk_type: str
    create: bool

    def to_hcl(self):
        return to_hcl(
            {
                "machine_type": self.machine_type,
                "disk_size": self.disk_size,
                "disk_type": self.disk_type,
                "create": self.create,
            }
        )

    @staticmethod
    def from_hcl(string: str) -> VirtualMachine:
        return VirtualMachine.from_dict(from_hcl(string))

    @staticmethod
    def from_dict(dict_: Dict) -> VirtualMachine:
        """Constructs a VirtualMachine instance from a dictionary.

        :param dict_: the dictionary.
        :return: the VirtualMachine instance.
        """

        machine_type = dict_.get("machine_type")
        disk_size = dict_.get("disk_size")
        disk_type = dict_.get("disk_type")
        create = str(dict_.get("create")).lower() == "true"
        return VirtualMachine(machine_type, disk_size, disk_type, create)


@dataclass
class ElasticSearch:
    """The elasticsearch settings for the Observatory Platform API.

    Attributes:
        host: the address of the elasticsearch host
        api_key: the api key to use the elasticsearch API.
    """

    host: str
    api_key: str

    def to_hcl(self):
        return to_hcl(
            {
                "host": self.host,
                "api_key": self.api_key,
            }
        )

    @staticmethod
    def from_dict(dict_: Dict) -> ElasticSearch:
        """Constructs a CloudSqlDatabase instance from a dictionary.

        :param dict_: the dictionary.
        :return: the CloudSqlDatabase instance.
        """

        host = dict_.get("host")
        api_key = dict_.get("api_key")
        return ElasticSearch(host, api_key)


@dataclass
class Api:
    """The API domain name for the Observatory Platform API.

    Attributes:
        domain_name: the custom domain name of the API
        subdomain: the subdomain of the API, can be either based on the google project id or the environment. When
        based on the environment, there is no subdomain for the production environment.
        api_image: the path to the Observatory API image on Google Cloud Artifact Registry.
        er_image: the path to the Google Cloud Endpoints Runtime image.
    """

    domain_name: str
    subdomain: str
    api_image: str
    er_image: str = "gcr.io/endpoints-release/endpoints-runtime-serverless:2"

    def to_hcl(self):
        return to_hcl(
            {
                "domain_name": self.domain_name,
                "subdomain": self.subdomain,
                "api_image": self.api_image,
                "er_image": self.er_image,
            }
        )

    @staticmethod
    def from_dict(dict_: Dict) -> Api:
        """Constructs a CloudSqlDatabase instance from a dictionary.

        :param dict_: the dictionary.
        :return: the CloudSqlDatabase instance.
        """

        domain_name = dict_.get("domain_name")
        subdomain = dict_.get("subdomain")
        api_image = dict_.get("api_image")
        er_image = dict_.get("er_image", Api.er_image)
        return Api(domain_name, subdomain, api_image, er_image=er_image)


def is_base64(text: bytes) -> bool:
    """Check if the string is base64.
    :param text: Text to check.
    :return: Whether it is base64.
    """

    try:
        base64.decodebytes(text)
    except:
        return False

    return True


def is_secret_key(key: str) -> Tuple[bool, Union[str, None]]:
    """Check if the Airflow Flask webserver secret key is valid.
    :param key: Key to check.
    :return: Validity, and an error message if not valid.
    """

    key_bytes = bytes(key, "utf-8")
    message = None

    key_length = len(key_bytes)
    if key_length < 16:
        message = f"Secret key should be length >=16, but is length {key_length}."
        return False, message

    return True, message


def is_fernet_key(key: str) -> Tuple[bool, Union[str, None]]:
    """Check if the Fernet key is valid.
    :param key: Key to check.
    :return: Validity, and an error message if not valid.
    """

    key_bytes = bytes(key, "utf-8")

    try:
        decoded_key = base64.urlsafe_b64decode(key_bytes)
    except:
        message = f"Key {key} could not be urlsafe b64decoded."
        return False, message

    key_length = len(decoded_key)
    if key_length != 32:
        message = f"Decoded Fernet key should be length 32, but is length {key_length}."
        return False, message

    message = None
    return True, message


def check_schema_field_fernet_key(field: str, value: str, error: Callable):
    """
    :param field: Field name.
    :param value: Field value.
    :param error: Error handler passed in by Cerberus.
    """

    valid, message = is_fernet_key(value)

    if not valid:
        error(field, f"is not a valid Fernet key. Reason: {message}")


def check_schema_field_secret_key(field: str, value: str, error: Callable):
    """
    :param field: Field name.
    :param value: Field value.
    :param error: Error handler passed in by Cerberus.
    """

    valid, message = is_secret_key(value)

    if not valid:
        error(field, f"is not a valid secret key. Reason: {message}")


def customise_pointer(field, value, error):
    """Throw an error when a field contains the value ' <--' which means that the user should customise the
    value in the config file.

    :param field: the field.
    :param value: the value.
    :param error: ?
    :return: None.
    """

    if isinstance(value, str) and value.endswith(" <--"):
        error(field, "Customise value ending with ' <--'")


class ObservatoryConfigValidator(Validator):
    """Custom config Validator"""

    def _validate_google_application_credentials(self, google_application_credentials, field, value):
        """Validate that the Google Application Credentials file exists.
        The rule's arguments are validated against this schema: {'type': 'boolean'}
        """
        if (
            google_application_credentials
            and value is not None
            and isinstance(value, str)
            and not os.path.isfile(value)
        ):
            self._error(
                field,
                f"the file {value} does not exist. See "
                f"https://cloud.google.com/docs/authentication/getting-started for instructions on "
                f"how to create a service account and save the JSON key to your workstation.",
            )


@dataclass
class ValidationError:
    """A validation error found when parsing a config file.

    Attributes:
        key: the key in the config file.
        value: the error.
    """

    key: str
    value: Any


class ObservatoryConfig:
    def __init__(
        self,
        backend: Backend = None,
        observatory: Observatory = None,
        google_cloud: GoogleCloud = None,
        terraform: Terraform = None,
        airflow_variables: List[AirflowVariable] = None,
        airflow_connections: List[AirflowConnection] = None,
        workflows_projects: List[WorkflowsProject] = None,
        validator: ObservatoryConfigValidator = None,
    ):
        """Create an ObservatoryConfig instance.

        :param backend: the backend config.
        :param observatory: the Observatory config.
        :param google_cloud: the Google Cloud config.
        :param terraform: the Terraform config.
        :param airflow_variables: a list of Airflow variables.
        :param airflow_connections: a list of Airflow connections.
        :param workflows_projects: a list of DAGs projects.
        :param validator: an ObservatoryConfigValidator instance.
        """

        self.backend = backend if backend is not None else Backend()
        self.observatory = observatory if observatory is not None else Observatory()
        self.google_cloud = google_cloud
        self.terraform = terraform

        self.airflow_variables = airflow_variables
        if airflow_variables is None:
            self.airflow_variables = []

        self.airflow_connections = airflow_connections
        if airflow_variables is None:
            self.airflow_connections = []

        self.workflows_projects = workflows_projects
        if workflows_projects is None:
            self.workflows_projects = []

        self.validator = validator

        self.schema = make_schema(self.backend.type)

    @property
    def is_valid(self) -> bool:
        """Checks whether the config is valid or not.

        :return: whether the config is valid or not.
        """

        return self.validator is None or not len(self.validator._errors)

    @property
    def errors(self) -> List[ValidationError]:
        """Returns a list of ValidationError instances that were created when parsing the config file.

        :return: the list of ValidationError instances.
        """

        errors = []
        for key, values in self.validator.errors.items():
            for value in values:
                if type(value) is dict:
                    for nested_key, nested_value in value.items():
                        errors.append(ValidationError(f"{key}.{nested_key}", *nested_value))
                else:
                    errors.append(ValidationError(key, *values))

        return errors

    @property
    def python_packages(self) -> List[PythonPackage]:
        """Make a list of Python Packages to build or include in the observatory.
        :return: the list of Python packages.
        """

        # observatory-api should be installed first so that observatory-platform install doesn't try to install
        # observatory-api from PyPI
        packages = [
            PythonPackage(
                name="observatory-api",
                type=self.observatory.api_package_type,
                host_package=self.observatory.api_package,
                docker_package=os.path.basename(self.observatory.api_package),
            ),
            PythonPackage(
                name="observatory-platform",
                type=self.observatory.package_type,
                host_package=self.observatory.package,
                docker_package=os.path.basename(self.observatory.package),
            ),
        ]

        for project in self.workflows_projects:
            packages.append(
                PythonPackage(
                    name=project.package_name,
                    type=project.package_type,
                    host_package=project.package,
                    docker_package=os.path.basename(project.package),
                )
            )

        return packages

    @property
    def dags_module_names(self):
        """Returns a list of DAG project module names.
        :return: the list of DAG project module names.
        """

        return f"'{str(json.dumps([project.dags_module for project in self.workflows_projects]))}'"

    def make_airflow_variables(self) -> List[AirflowVariable]:
        """Make all airflow variables for the observatory platform.

        :return: a list of AirflowVariable objects.
        """

        # Create airflow variables from fixed config file values
        variables = [AirflowVariable(AirflowVars.ENVIRONMENT, self.backend.environment.value)]

        if self.google_cloud is not None:
            if self.google_cloud.project_id is not None:
                variables.append(AirflowVariable(AirflowVars.PROJECT_ID, self.google_cloud.project_id))

            if self.google_cloud.data_location:
                variables.append(AirflowVariable(AirflowVars.DATA_LOCATION, self.google_cloud.data_location))

            # Create airflow variables from bucket names
            for bucket in self.google_cloud.buckets:
                variables.append(AirflowVariable(bucket.id, bucket.name))

        if self.terraform is not None:
            if self.terraform.organization is not None:
                variables.append(AirflowVariable(AirflowVars.TERRAFORM_ORGANIZATION, self.terraform.organization))

        # Add user defined variables to list
        variables += self.airflow_variables

        return variables

    @staticmethod
    def _parse_fields(
        dict_: Dict,
    ) -> Tuple[
        Backend,
        Observatory,
        GoogleCloud,
        Terraform,
        List[AirflowVariable],
        List[AirflowConnection],
        List[WorkflowsProject],
    ]:
        backend = Backend.from_dict(dict_.get("backend", dict()))
        observatory = Observatory.from_dict(dict_.get("observatory", dict()))
        google_cloud = GoogleCloud.from_dict(dict_.get("google_cloud", dict()))
        terraform = Terraform.from_dict(dict_.get("terraform", dict()))
        airflow_variables = AirflowVariable.parse_airflow_variables(dict_.get("airflow_variables", dict()))
        airflow_connections = AirflowConnection.parse_airflow_connections(dict_.get("airflow_connections", dict()))
        workflows_projects = WorkflowsProject.parse_workflows_projects(dict_.get("workflows_projects", list()))

        return backend, observatory, google_cloud, terraform, airflow_variables, airflow_connections, workflows_projects

    @classmethod
    def from_dict(cls, dict_: Dict) -> ObservatoryConfig:
        """Constructs an ObservatoryConfig instance from a dictionary.

        If the dictionary is invalid, then an ObservatoryConfig instance will be returned with no properties set,
        except for the validator, which contains validation errors.

        :param dict_: the input dictionary.
        :return: the ObservatoryConfig instance.
        """

        schema = make_schema(BackendType.local)
        validator = ObservatoryConfigValidator()
        is_valid = validator.validate(dict_, schema)

        if is_valid:
            (
                backend,
                observatory,
                google_cloud,
                terraform,
                airflow_variables,
                airflow_connections,
                workflows_projects,
            ) = ObservatoryConfig._parse_fields(dict_)

            return ObservatoryConfig(
                backend,
                observatory,
                google_cloud=google_cloud,
                terraform=terraform,
                airflow_variables=airflow_variables,
                airflow_connections=airflow_connections,
                workflows_projects=workflows_projects,
                validator=validator,
            )
        else:
            return ObservatoryConfig(validator=validator)

    @classmethod
    def load(cls, path: str):
        """Load a configuration file.

        :return: the ObservatoryConfig instance (or a subclass of ObservatoryConfig)
        """

        dict_ = dict()

        try:
            with open(path, "r") as f:
                dict_ = yaml.safe_load(f)
        except yaml.YAMLError:
            print(f"Error parsing {path}")
        except FileNotFoundError:
            print(f"No such file or directory: {path}")
        except cerberus.validator.DocumentError as e:
            print(f"cerberus.validator.DocumentError: {e}")

        return cls.from_dict(dict_)

    def get_requirement_string(self, section: str) -> str:
        """Query the schema to see whether a section is required.

        :param section: Section to query.
        :return: String indicating whether the section is required or optional.
        """

        if self.schema[section]["required"]:
            return "Required"

        return "Optional"

    def save(self, path: str):
        """Save the observatory configuration parameters to a config file.

        :param path: Configuration file path.
        """

        with open(path, "w") as f:
            self.save_backend(f)
            self.save_observatory(f)
            self.save_terraform(f)
            self.save_google_cloud(f)
            self.save_airflow_connections(f)
            self.save_airflow_variables(f)
            self.save_workflows_projects(f)

    def save_backend(self, f: TextIO):
        """Write the backend configuration section to the config file.

        :param f: File object for the config file.
        """

        requirement = self.get_requirement_string("backend")
        f.write(
            (
                f"# [{requirement}] Backend settings.\n"
                "# Backend options are: local, terraform.\n"
                "# Environment options are: develop, staging, production.\n"
            )
        )
        lines = ObserveratoryConfigString.backend(self.backend)
        f.writelines(lines)
        f.write("\n")

    def save_observatory(self, f: TextIO):
        """Write the Observatory configuration section to the config file.

        :param f: File object for the config file.
        """

        requirement = self.get_requirement_string("observatory")
        f.write(
            (
                f"# [{requirement}] Observatory settings\n"
                "# If you did not supply your own Fernet and secret keys, then those fields are autogenerated.\n"
                "# Passwords are in plaintext.\n"
                "# observatory_home is where the observatory metadata is stored.\n"
            )
        )

        lines = ObserveratoryConfigString.observatory(self.observatory)
        f.writelines(lines)
        f.write("\n")

    def save_terraform(self, f: TextIO):
        """Write the Terraform configuration section to the config file.

        :param f: File object for the config file.
        """

        requirement = self.get_requirement_string("terraform")
        f.write(f"# [{requirement}] Terraform settings\n")

        lines = ObserveratoryConfigString.terraform(self.terraform)
        output = map(comment, lines) if self.terraform is None and requirement == "Optional" else lines

        f.writelines(output)
        f.write("\n")

    def save_google_cloud(self, f: TextIO):
        """Write the Google Cloud configuration section to the config file.

        :param f: File object for the config file.
        """

        requirement = self.get_requirement_string("google_cloud")
        f.write(
            (
                f"# [{requirement}] Google Cloud settings\n"
                "# If you use any Google Cloud service functions, you will need to configure this.\n"
            )
        )

        lines = ObserveratoryConfigString.google_cloud(google_cloud=self.google_cloud, backend=self.backend)
        output = map(comment, lines) if self.google_cloud is None and requirement == "Optional" else lines

        f.writelines(output)
        f.write("\n")

    def save_airflow_connections(self, f: TextIO):
        """Write the Airflow connections configuration section to the config file.

        :param f: File object for the config file.
        """

        requirement = self.get_requirement_string("airflow_connections")
        f.write(f"# [{requirement}] User defined Apache Airflow Connections:\n")

        lines = ObserveratoryConfigString.airflow_connections(airflow_connections=self.airflow_connections)
        output = map(comment, lines) if len(self.airflow_connections) == 0 and requirement == "Optional" else lines

        f.writelines(output)
        f.write("\n")

    def save_airflow_variables(self, f: TextIO):
        """Write the Airflow variables configuration section to the config file.

        :param f: File object for the config file.
        """

        requirement = self.get_requirement_string("airflow_variables")
        f.write(f"# [{requirement}] User defined Apache Airflow variables:\n")

        lines = ObserveratoryConfigString.airflow_variables(airflow_variables=self.airflow_variables)
        output = map(comment, lines) if len(self.airflow_variables) == 0 and requirement == "Optional" else lines

        f.writelines(output)
        f.write("\n")

    def save_workflows_projects(self, f: TextIO):
        """Write the DAGs projects configuration section to the config file.

        :param f: File object for the config file.
        """

        requirement = self.get_requirement_string("workflows_projects")
        f.write(f"# [{requirement}] User defined Observatory DAGs projects:\n")

        lines = ObserveratoryConfigString.workflows_projects(workflows_projects=self.workflows_projects)
        output = map(comment, lines) if len(self.workflows_projects) == 0 and requirement == "Optional" else lines

        f.writelines(output)
        f.write("\n")


class TerraformConfig(ObservatoryConfig):
    WORKSPACE_PREFIX = "observatory-"

    def __init__(
        self,
        backend: Backend = None,
        observatory: Observatory = None,
        google_cloud: GoogleCloud = None,
        terraform: Terraform = None,
        airflow_variables: List[AirflowVariable] = None,
        airflow_connections: List[AirflowConnection] = None,
        workflows_projects: List[WorkflowsProject] = None,
        cloud_sql_database: CloudSqlDatabase = None,
        airflow_main_vm: VirtualMachine = None,
        airflow_worker_vm: VirtualMachine = None,
        elasticsearch: ElasticSearch = None,
        api: Api = None,
        validator: ObservatoryConfigValidator = None,
    ):
        """Create a TerraformConfig instance.

        :param backend: the backend config.
        :param observatory: the Observatory config.
        :param google_cloud: the Google Cloud config.
        :param terraform: the Terraform config.
        :param airflow_variables: a list of Airflow variables.
        :param airflow_connections: a list of Airflow connections.
        :param workflows_projects: a list of DAGs projects.
        :param cloud_sql_database: a Google Cloud SQL database config.
        :param airflow_main_vm: the Airflow Main VM config.
        :param airflow_worker_vm: the Airflow Worker VM config.
        :param validator: an ObservatoryConfigValidator instance.
        """

        if backend is None:
            backend = Backend(type=BackendType.terraform)

        super().__init__(
            backend=backend,
            observatory=observatory,
            google_cloud=google_cloud,
            terraform=terraform,
            airflow_variables=airflow_variables,
            airflow_connections=airflow_connections,
            workflows_projects=workflows_projects,
            validator=validator,
        )
        self.cloud_sql_database = cloud_sql_database
        self.airflow_main_vm = airflow_main_vm
        self.airflow_worker_vm = airflow_worker_vm
        self.elasticsearch = elasticsearch
        self.api = api

    @property
    def terraform_workspace_id(self):
        """The Terraform workspace id.

        :return: the terraform workspace id.
        """

        return TerraformConfig.WORKSPACE_PREFIX + self.backend.environment.value

    def make_airflow_variables(self) -> List[AirflowVariable]:
        """Make all airflow variables for the Observatory Platform.

        :return: a list of AirflowVariable objects.
        """

        # Create airflow variables from fixed config file values
        variables = [AirflowVariable(AirflowVars.ENVIRONMENT, self.backend.environment.value)]

        if self.google_cloud.project_id is not None:
            variables.append(AirflowVariable(AirflowVars.PROJECT_ID, self.google_cloud.project_id))
            variables.append(AirflowVariable(AirflowVars.DOWNLOAD_BUCKET, f"{self.google_cloud.project_id}-download"))
            variables.append(AirflowVariable(AirflowVars.TRANSFORM_BUCKET, f"{self.google_cloud.project_id}-transform"))

        if self.google_cloud.data_location:
            variables.append(AirflowVariable(AirflowVars.DATA_LOCATION, self.google_cloud.data_location))

        if self.terraform.organization is not None:
            variables.append(AirflowVariable(AirflowVars.TERRAFORM_ORGANIZATION, self.terraform.organization))

        variables.append(
            AirflowVariable(
                AirflowVars.DAGS_MODULE_NAMES, json.dumps([proj.dags_module for proj in self.workflows_projects])
            )
        )

        # Add user defined variables to list
        variables += self.airflow_variables

        return variables

    def terraform_variables(self) -> List[TerraformVariable]:
        """Create a list of TerraformVariable instances from the Terraform Config.

        :return: a list of TerraformVariable instances.
        """

        sensitive = True
        return [
            TerraformVariable("environment", self.backend.environment.value),
            TerraformVariable("observatory", self.observatory.to_hcl(), sensitive=sensitive, hcl=True),
            TerraformVariable("google_cloud", self.google_cloud.to_hcl(), sensitive=sensitive, hcl=True),
            TerraformVariable("cloud_sql_database", self.cloud_sql_database.to_hcl(), hcl=True),
            TerraformVariable("airflow_main_vm", self.airflow_main_vm.to_hcl(), hcl=True),
            TerraformVariable("airflow_worker_vm", self.airflow_worker_vm.to_hcl(), hcl=True),
            TerraformVariable(
                "airflow_variables", list_to_hcl(self.make_airflow_variables()), hcl=True, sensitive=False
            ),
            TerraformVariable(
                "airflow_connections", list_to_hcl(self.airflow_connections), hcl=True, sensitive=sensitive
            ),
            TerraformVariable("elasticsearch", self.elasticsearch.to_hcl(), sensitive=sensitive, hcl=True),
            TerraformVariable("api", self.api.to_hcl(), hcl=True),
        ]

    @classmethod
    def from_dict(cls, dict_: Dict) -> TerraformConfig:
        """Make an TerraformConfig instance from a dictionary.

        If the dictionary is invalid, then an ObservatoryConfig instance will be returned with no properties set,
        except for the validator, which contains validation errors.

        :param dict_: the input dictionary that has been read via yaml.safe_load.
        :return: the TerraformConfig instance.
        """

        schema = make_schema(BackendType.terraform)
        validator = ObservatoryConfigValidator()
        is_valid = validator.validate(dict_, schema)

        if is_valid:
            (
                backend,
                observatory,
                google_cloud,
                terraform,
                airflow_variables,
                airflow_connections,
                workflows_projects,
            ) = ObservatoryConfig._parse_fields(dict_)

            cloud_sql_database = CloudSqlDatabase.from_dict(dict_.get("cloud_sql_database", dict()))
            airflow_main_vm = VirtualMachine.from_dict(dict_.get("airflow_main_vm", dict()))
            airflow_worker_vm = VirtualMachine.from_dict(dict_.get("airflow_worker_vm", dict()))
            elasticsearch = ElasticSearch.from_dict(dict_.get("elasticsearch", dict()))
            api = Api.from_dict(dict_.get("api", dict()))

            return TerraformConfig(
                backend,
                observatory,
                google_cloud=google_cloud,
                terraform=terraform,
                airflow_variables=airflow_variables,
                airflow_connections=airflow_connections,
                workflows_projects=workflows_projects,
                cloud_sql_database=cloud_sql_database,
                airflow_main_vm=airflow_main_vm,
                airflow_worker_vm=airflow_worker_vm,
                elasticsearch=elasticsearch,
                api=api,
                validator=validator,
            )
        else:
            return TerraformConfig(validator=validator)

    def save(self, path: str):
        """Save the configuration to a config file in YAML format.

        :param path: Config file path.
        """

        # Save common config
        super().save(path)

        # Save Terraform specific sections
        with open(path, "a") as f:
            self.save_cloud_sql_database(f)
            self.save_airflow_main_vm(f)
            self.save_airflow_worker_vm(f)
            self.save_elasticsearch(f)
            self.save_api(f)

    def save_cloud_sql_database(self, f: TextIO):
        """Write the cloud SQL database configuration section to the config file.

        :param f: File object for the config file.
        """

        requirement = self.get_requirement_string("cloud_sql_database")
        f.write(f"# [{requirement}] Google Cloud CloudSQL database settings\n")
        lines = ObserveratoryConfigString.cloud_sql_database(self.cloud_sql_database)
        f.writelines(lines)
        f.write("\n")

    def save_airflow_main_vm(self, f: TextIO):
        """Write the Airflow main VM configuration section to the config file.

        :param f: File object for the config file.
        """

        requirement = self.get_requirement_string("airflow_main_vm")
        f.write(f"# [{requirement}] Settings for the main VM that runs the Airflow cheduler and webserver\n")
        lines = ObserveratoryConfigString.airflow_main_vm(self.airflow_main_vm)
        f.writelines(lines)
        f.write("\n")

    def save_airflow_worker_vm(self, f: TextIO):
        """Write the Airflow worker VM configuration section to the config file.

        :param f: File object for the config file.
        """

        requirement = self.get_requirement_string("airflow_worker_vm")
        f.write(f"# [{requirement}] Settings for the weekly on-demand VM that runs arge tasks\n")
        lines = ObserveratoryConfigString.airflow_worker_vm(self.airflow_worker_vm)
        f.writelines(lines)
        f.write("\n")

    def save_elasticsearch(self, f: TextIO):
        """Write the ElasticSearch configuration section to the config file.

        :param f: File object for the config file.
        """

        requirement = self.get_requirement_string("elasticsearch")
        f.write(f"# [{requirement}] Elasticsearch settings\n")
        lines = ObserveratoryConfigString.elasticsearch(self.elasticsearch)
        f.writelines(lines)
        f.write("\n")

    def save_api(self, f: TextIO):
        """Write the API configuration section to the config file.

        :param f: File object for the config file.
        """

        requirement = self.get_requirement_string("api")
        f.write(
            (
                f"# [{requirement}] API settings\n"
                "# The subdomain type determines what kind of subdomains will be created for access."
                "# The options are: project_id, environment"
                "# If project_id is selected, the project_id will be used for API access for that project, e.g., project_id.domain_name"
                "# If environment is selected, the environment parameter will be used, e.g., production.domain_name\n"
            )
        )
        lines = ObserveratoryConfigString.api(self.api)
        f.writelines(lines)
        f.write("\n")


Config = Union[ObservatoryConfig, TerraformConfig]


def make_schema(backend_type: BackendType) -> Dict:
    """Make a schema for an Observatory or Terraform config file.

    :param backend_type: the type of backend, local or terraform.
    :return: a dictionary containing the schema.
    """

    schema = dict()
    is_backend_terraform = backend_type == BackendType.terraform

    # Backend settings
    schema["backend"] = {
        "required": True,
        "type": "dict",
        "schema": {
            "type": {"required": True, "type": "string", "allowed": [backend_type.value]},
            "environment": {"required": True, "type": "string", "allowed": ["develop", "staging", "production"]},
        },
    }

    # Terraform settings
    schema["terraform"] = {
        "required": is_backend_terraform,
        "type": "dict",
        "schema": {"organization": {"required": True, "type": "string", "check_with": customise_pointer}},
    }

    # Google Cloud settings
    schema["google_cloud"] = {
        "required": is_backend_terraform,
        "type": "dict",
        "schema": {
            "project_id": {"required": is_backend_terraform, "type": "string", "check_with": customise_pointer},
            "credentials": {
                "required": is_backend_terraform,
                "type": "string",
                "check_with": customise_pointer,
                "google_application_credentials": True,
            },
            "region": {
                "required": is_backend_terraform,
                "type": "string",
                "regex": r"^\w+\-\w+\d+$",
                "check_with": customise_pointer,
            },
            "zone": {
                "required": is_backend_terraform,
                "type": "string",
                "regex": r"^\w+\-\w+\d+\-[a-z]{1}$",
                "check_with": customise_pointer,
            },
            "data_location": {"required": is_backend_terraform, "type": "string", "check_with": customise_pointer},
        },
    }

    if not is_backend_terraform:
        schema["google_cloud"]["schema"]["buckets"] = {
            "required": False,
            "type": "dict",
            "keysrules": {"type": "string"},
            "valuesrules": {"type": "string"},
        }

    # Observatory settings
    package_types = ["editable", "sdist", "pypi"]
    schema["observatory"] = {
        "required": True,
        "type": "dict",
        "schema": {
            "package": {"required": True, "type": "string"},
            "package_type": {"required": True, "type": "string", "allowed": package_types},
            "airflow_fernet_key": {"required": True, "type": "string", "check_with": check_schema_field_fernet_key},
            "airflow_secret_key": {"required": True, "type": "string", "check_with": check_schema_field_secret_key},
            "airflow_ui_user_password": {"required": is_backend_terraform, "type": "string"},
            "airflow_ui_user_email": {"required": is_backend_terraform, "type": "string"},
            "observatory_home": {"required": False, "type": "string"},
            "postgres_password": {"required": is_backend_terraform, "type": "string"},
            "redis_port": {"required": False, "type": "integer"},
            "flower_ui_port": {"required": False, "type": "integer"},
            "airflow_ui_port": {"required": False, "type": "integer"},
            "elastic_port": {"required": False, "type": "integer"},
            "kibana_port": {"required": False, "type": "integer"},
            "docker_network_name": {"required": False, "type": "string"},
            "docker_network_is_external": {"required": False, "type": "boolean"},
            "docker_compose_project_name": {"required": False, "type": "string"},
            "enable_elk": {"required": False, "type": "boolean"},
            "api_package": {"required": False, "type": "string"},
            "api_package_type": {"required": False, "type": "string", "allowed": package_types},
            "api_port": {"required": False, "type": "integer"},
        },
    }

    # Database settings
    if is_backend_terraform:
        schema["cloud_sql_database"] = {
            "required": True,
            "type": "dict",
            "schema": {
                "tier": {"required": True, "type": "string"},
                "backup_start_time": {"required": True, "type": "string", "regex": r"^\d{2}:\d{2}$"},
            },
        }

    # VM schema
    vm_schema = {
        "required": True,
        "type": "dict",
        "schema": {
            "machine_type": {
                "required": True,
                "type": "string",
            },
            "disk_size": {"required": True, "type": "integer", "min": 1},
            "disk_type": {"required": True, "type": "string", "allowed": ["pd-standard", "pd-ssd"]},
            "create": {"required": True, "type": "boolean"},
        },
    }

    # Airflow main and worker VM
    if is_backend_terraform:
        schema["airflow_main_vm"] = vm_schema
        schema["airflow_worker_vm"] = vm_schema

    # Key value string pair schema
    key_val_schema = {
        "required": False,
        "type": "dict",
        "keysrules": {"type": "string"},
        "valuesrules": {"type": "string"},
    }

    # Airflow variables
    schema["airflow_variables"] = key_val_schema

    # Airflow connections
    schema["airflow_connections"] = {
        "required": False,
        "type": "dict",
        "keysrules": {"type": "string"},
        "valuesrules": {"type": "string", "regex": r"\S*:\/\/\S*:\S*@\S*$|google-cloud-platform:\/\/\S*$"},
    }

    # Dags projects
    schema["workflows_projects"] = {
        "required": False,
        "type": "list",
        "schema": {
            "type": "dict",
            "schema": {
                "package_name": {
                    "required": True,
                    "type": "string",
                },
                "package": {"required": True, "type": "string"},
                "package_type": {"required": True, "type": "string", "allowed": package_types},
                "dags_module": {
                    "required": True,
                    "type": "string",
                },
            },
        },
    }

    if is_backend_terraform:
        schema["elasticsearch"] = {
            "required": True,
            "type": "dict",
            "schema": {
                "host": {"required": True, "type": "string"},
                "api_key": {
                    "required": True,
                    "type": "string",
                },
            },
        }

        schema["api"] = {
            "required": True,
            "type": "dict",
            "schema": {
                "domain_name": {"required": True, "type": "string"},
                "subdomain": {"required": True, "type": "string", "allowed": ["project_id", "environment"]},
                "api_image": {"required": True, "type": "string"},
                "er_image": {"required": False, "type": "string"},
            },
        }

    return schema


class ObserveratoryConfigString:
    """This class contains methods to construct config file sections."""

    @staticmethod
    def backend(backend: Backend) -> List[str]:
        """Constructs the backend section string.

        :param backend: Backend configuration object.
        :return: List of strings for the section, including the section heading."
        """

        lines = [
            "backend:\n",
            indent(f"type: {backend.type.name}\n", INDENT1),
            indent(f"environment: {backend.environment.name}\n", INDENT1),
        ]

        return lines

    @staticmethod
    def observatory(observatory: Observatory) -> List[str]:
        """Constructs the observatory section string.

        :param observatory: Observatory configuration object.
        :return: List of strings for the section, including the section heading."
        """

        lines = [
            "observatory:\n",
            indent(f"package: {observatory.package}\n", INDENT1),
            indent(f"package_type: {observatory.package_type}\n", INDENT1),
            indent(f"airflow_fernet_key: {observatory.airflow_fernet_key}\n", INDENT1),
            indent(f"airflow_secret_key: {observatory.airflow_secret_key}\n", INDENT1),
            indent(f"airflow_ui_user_email: {observatory.airflow_ui_user_email}\n", INDENT1),
            indent(f"airflow_ui_user_password: {observatory.airflow_ui_user_password}\n", INDENT1),
            indent(f"observatory_home: {observatory.observatory_home}\n", INDENT1),
            indent(f"postgres_password: {observatory.postgres_password}\n", INDENT1),
            indent(f"redis_port: {observatory.redis_port}\n", INDENT1),
            indent(f"flower_ui_port: {observatory.flower_ui_port}\n", INDENT1),
            indent(f"airflow_ui_port: {observatory.airflow_ui_port}\n", INDENT1),
            indent(f"elastic_port: {observatory.elastic_port}\n", INDENT1),
            indent(f"kibana_port: {observatory.kibana_port}\n", INDENT1),
            indent(f"docker_network_name: {observatory.docker_network_name}\n", INDENT1),
            indent(f"docker_network_is_external: {observatory.docker_network_is_external}\n", INDENT1),
            indent(f"docker_compose_project_name: {observatory.docker_compose_project_name}\n", INDENT1),
            indent(f"enable_elk: {observatory.enable_elk}\n", INDENT1),
            indent(f"api_package: {observatory.api_package}\n", INDENT1),
            indent(f"api_package_type: {observatory.api_package_type}\n", INDENT1),
            indent(f"api_port: {observatory.api_port}\n", INDENT1),
        ]

        return lines

    @staticmethod
    def terraform(terraform: Terraform) -> List[str]:
        """Constructs the terraform section string.

        :param observatory: Terraform configuration object.
        :return: List of strings for the section, including the section heading."
        """

        if terraform is None:
            terraform = Terraform(organization="my-terraform-org-name")

        lines = [
            "terraform:\n",
            indent(f"organization: {terraform.organization}\n", INDENT1),
        ]

        return lines

    @staticmethod
    def google_cloud(*, google_cloud: GoogleCloud, backend: Backend) -> List[str]:
        """Constructs the Google Cloud section string.

        :param google_cloud: Google Cloud configuration object.
        :param backend: Backend configuration object.
        :return: List of strings for the section, including the section heading."
        """

        if google_cloud is None:
            google_cloud = GoogleCloud(
                project_id="my-gcp-id",
                credentials="/path/to/credentials.json",
                data_location="us",
                region="us-west1",
                zone="us-west1-a",
                buckets=[
                    CloudStorageBucket(
                        id="download_bucket",
                        name="my-download-bucket-name",
                    ),
                    CloudStorageBucket(
                        id="transform_bucket",
                        name="my-transform-bucket-name",
                    ),
                ],
            )

        lines = [
            "google_cloud:\n",
            indent(f"project_id: {google_cloud.project_id}\n", INDENT1),
            indent(f"credentials: {google_cloud.credentials}\n", INDENT1),
            indent(f"data_location: {google_cloud.data_location}\n", INDENT1),
        ]

        # Is region and zone something we should be putting in the local config too?
        if backend.type == BackendType.terraform:
            lines.append(indent(f"region: {google_cloud.region}\n", INDENT1))
            lines.append(indent(f"zone: {google_cloud.zone}\n", INDENT1))
        else:
            lines.append(indent("buckets:\n", INDENT1))
            for bucket in google_cloud.buckets:
                lines.append(indent(f"{bucket.id}: {bucket.name}\n", INDENT3))

        return lines

    @staticmethod
    def airflow_connections(airflow_connections: List[AirflowConnection]) -> List[str]:
        """Constructs the Airflow connections section string.

        :param airflow_connections: List of Airflow connection configuration objects.
        :return: List of strings for the section, including the section heading."
        """

        lines = ["airflow_connections:\n"]

        connections = airflow_connections.copy()
        if len(connections) == 0:
            connections.append(
                AirflowConnection(
                    name="my_connection",
                    value="http://my-username:my-password@",
                )
            )

        for conn in connections:
            lines.append(indent(f"{conn.name}: {conn.value}\n", INDENT1))

        return lines

    @staticmethod
    def airflow_variables(airflow_variables: List[AirflowVariable]) -> List[str]:
        """Constructs the Airflow variables section string.

        :param airflow_connections: List of Airflow variable configuration objects.
        :return: List of strings for the section, including the section heading."
        """

        lines = ["airflow_variables:\n"]

        variables = airflow_variables.copy()

        if len(variables) == 0:
            variables.append(
                AirflowVariable(
                    name="my_variable_name",
                    value="my-variable-value",
                )
            )

        for variable in variables:
            lines.append(indent(f"{variable.name}: {variable.value}\n", INDENT1))

        return lines

    @staticmethod
    def workflows_projects(*, workflows_projects: List[WorkflowsProject] = None) -> List[str]:
        """Constructs the DAGs projects section string.

        :param workflows_projects: List of DAGs project configuration objects.
        :return: List of strings for the section, including the section heading."
        """

        projects = workflows_projects.copy()

        if len(projects) == 0:
            projects.append(
                WorkflowsProject(
                    package_name="observatory-dags",
                    package="/path/to/dags_project",
                    package_type="editable",
                    dags_module="observatory.dags.dags",
                )
            )

        lines = ["workflows_projects:\n"]
        for project in projects:
            lines.append(indent(f"- package_name: {project.package_name}\n", INDENT1))
            lines.append(indent(f"package: {project.package}\n", INDENT3))
            lines.append(indent(f"package_type: {project.package_type}\n", INDENT3))
            lines.append(indent(f"dags_module: {project.dags_module}\n", INDENT3))

        return lines

    @staticmethod
    def cloud_sql_database(cloud_sql_database: CloudSqlDatabase) -> List[str]:
        """Constructs the cloud SQL database section string.

        :param cloud_sql_database: Cloud SQL configuration object.
        :return: List of strings for the section, including the section heading."
        """

        if cloud_sql_database is None:
            cloud_sql_database = CloudSqlDatabase(
                tier="db-custom-2-7680",
                backup_start_time="23:00",
            )

        lines = [
            "cloud_sql_database:\n",
            indent(f"tier: {cloud_sql_database.tier}\n", INDENT1),
            indent(f"backup_start_time: '{cloud_sql_database.backup_start_time}'\n", INDENT1),
        ]

        return lines

    @staticmethod
    def airflow_vm_lines_(*, vm: VirtualMachine, vm_type) -> List[str]:
        """Constructs the virtual machine section string.

        :param vm: Virtual machine configuration object.
        :param vm_type: Type of vm being configured.
        :return: List of strings for the section, including the section heading."
        """
        lines = [
            f"{vm_type}:\n",
            indent(f"machine_type: {vm.machine_type}\n", INDENT1),
            indent(f"disk_size: {vm.disk_size}\n", INDENT1),
            indent(f"disk_type: {vm.disk_type}\n", INDENT1),
            indent(f"create: {vm.create}\n", INDENT1),
        ]

        return lines

    @staticmethod
    def airflow_main_vm(vm: VirtualMachine) -> List[str]:
        """Constructs the main virtual machine section string.

        :param vm: Virtual machine configuration object.
        :return: List of strings for the section, including the section heading."
        """

        if vm is None:
            vm = VirtualMachine(
                machine_type="n2-standard-2",
                disk_size=50,
                disk_type="pd-ssd",
                create=True,
            )

        lines = ObserveratoryConfigString.airflow_vm_lines_(vm=vm, vm_type="airflow_main_vm")
        return lines

    @staticmethod
    def airflow_worker_vm(vm: VirtualMachine) -> List[str]:
        """Constructs the worker virtual machine section string.

        :param vm: Virtual machine configuration object.
        :return: List of strings for the section, including the section heading."
        """

        if vm is None:
            vm = VirtualMachine(
                machine_type="n1-standard-8",
                disk_size=3000,
                disk_type="pd-standard",
                create=False,
            )

        lines = ObserveratoryConfigString.airflow_vm_lines_(vm=vm, vm_type="airflow_worker_vm")
        return lines

    @staticmethod
    def elasticsearch(elasticsearch: ElasticSearch) -> List[str]:
        """Constructs the ElasticSearch section string.

        :param elasticsearch: Elastic search configuration object.
        :return: List of strings for the section, including the section heading."
        """

        if elasticsearch is None:
            elasticsearch = ElasticSearch(
                host="https://address.region.gcp.cloud.es.io:port",
                api_key="myapikey",
            )

        lines = [
            "elasticsearch:\n",
            indent(f"host: {elasticsearch.host}\n", INDENT1),
            indent(f"api_key: {elasticsearch.api_key}\n", INDENT1),
        ]

        return lines

    @staticmethod
    def api(api: Api) -> List[str]:
        """Constructs the Observatory API section string.

        :param api: API configuration object.
        :return: List of strings for the section, including the section heading."
        """

        if api is None:
            api = Api(
                domain_name="api.observatory.academy",
                subdomain="project_id",
                api_image="us-docker.pkg.dev/gcp-project-id/observatory-platform/observatory-api:latest",
            )

        lines = [
            "api:\n",
            indent(f"domain_name: {api.domain_name}\n", INDENT1),
            indent(f"subdomain: {api.subdomain}\n", INDENT1),
            indent(f"api_image: {api.api_image}\n", INDENT1),
            indent(f"er_image: {api.er_image}\n", INDENT1),
        ]

        return lines



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose


import distutils.dir_util
import os
import shutil
import subprocess
from subprocess import Popen
from typing import Tuple

from observatory.api.server.openapi_renderer import OpenApiRenderer
from observatory.platform.cli.cli_utils import indent, INDENT1
from observatory.platform.docker.platform_runner import PlatformRunner
from observatory.platform.observatory_config import TerraformConfig
from observatory.platform.utils.config_utils import module_file_path
from observatory.platform.utils.jinja2_utils import render_template
from observatory.platform.utils.proc_utils import stream_process


def copy_dir(source_path: str, destination_path: str, ignore):
    distutils.dir_util.copy_tree(source_path, destination_path)


class TerraformBuilder:
    def __init__(self, config: TerraformConfig, debug: bool = False):
        """Create a TerraformBuilder instance, which is used to build, start and stop an Observatory Platform instance.

        :param config: the TerraformConfig.
        :param debug: whether to print debug statements.
        """

        self.config = config
        self.api_package_path = module_file_path("observatory.api", nav_back_steps=-3)
        self.terraform_path = module_file_path("observatory.platform.terraform")
        self.api_path = module_file_path("observatory.api.server")
        self.debug = debug

        self.build_path = os.path.join(config.observatory.observatory_home, "build", "terraform")
        self.platform_runner = PlatformRunner(config=config, docker_build_path=os.path.join(self.build_path, "docker"))
        self.packages_build_path = os.path.join(self.build_path, "packages")
        self.terraform_build_path = os.path.join(self.build_path, "terraform")
        os.makedirs(self.packages_build_path, exist_ok=True)
        os.makedirs(self.terraform_build_path, exist_ok=True)

    @property
    def is_environment_valid(self) -> bool:
        """Return whether the environment for building the Packer image is valid.

        :return: whether the environment for building the Packer image is valid.
        """

        return all([self.packer_exe_path is not None])

    @property
    def packer_exe_path(self) -> str:
        """The path to the Packer executable.

        :return: the path or None.
        """

        return shutil.which("packer")

    @property
    def gcloud_exe_path(self) -> str:
        """The path to the Google Cloud SDK executable.

        :return: the path or None.
        """

        return shutil.which("gcloud")

    def make_files(self):
        # Clear terraform/packages path
        if os.path.exists(self.packages_build_path):
            shutil.rmtree(self.packages_build_path)
        os.makedirs(self.packages_build_path)

        ignore = shutil.ignore_patterns("__pycache__", "*.eggs", "*.egg-info")

        # Copy local packages
        for package in self.config.python_packages:
            if package.type == "editable":
                destination_path = os.path.join(self.packages_build_path, package.name)
                copy_dir(package.host_package, destination_path, ignore)

        # Copy terraform files into build/terraform: ignore jinja2 templates
        copy_dir(self.terraform_path, self.terraform_build_path, shutil.ignore_patterns("*.jinja2", "__pycache__"))

        # Make startup scripts
        self.make_startup_script(True, "startup-main.tpl")
        self.make_startup_script(False, "startup-worker.tpl")

        # Make OpenAPI specification
        self.make_open_api_template()

    def make_startup_script(self, is_airflow_main_vm: bool, file_name: str):
        # Load and render template
        template_path = os.path.join(self.terraform_path, "startup.tpl.jinja2")
        render = render_template(template_path, is_airflow_main_vm=is_airflow_main_vm)

        # Save file
        output_path = os.path.join(self.terraform_build_path, file_name)
        with open(output_path, "w") as f:
            f.write(render)

    def make_open_api_template(self):
        # Load and render template
        specification_path = os.path.join(self.api_path, "openapi.yaml.jinja2")
        renderer = OpenApiRenderer(specification_path, cloud_endpoints=True)
        render = renderer.render()

        # Save file
        output_path = os.path.join(self.terraform_build_path, "openapi.yaml.tpl")
        with open(output_path, "w") as f:
            f.write(render)

    def build_terraform(self):
        """Build the Observatory Platform Terraform files.

        :return: None.
        """

        self.make_files()
        self.platform_runner.make_files()

    def build_image(self) -> Tuple[str, str, int]:
        """Build the Observatory Platform Google Compute image with Packer.

        :return: output and error stream results and proc return code.
        """

        # Make Terraform files
        self.build_terraform()

        # Load template
        template_vars = {
            "credentials_file": self.config.google_cloud.credentials,
            "project_id": self.config.google_cloud.project_id,
            "zone": self.config.google_cloud.zone,
            "environment": self.config.backend.environment.value,
        }
        variables = []
        for key, val in template_vars.items():
            variables.append("-var")
            variables.append(f"{key}={val}")

        # Build the containers first
        args = ["packer", "build"] + variables + ["-force", "observatory-image.json"]

        if self.debug:
            print("Executing subprocess:")
            print(indent(f"Command: {subprocess.list2cmdline(args)}", INDENT1))
            print(indent(f"Cwd: {self.terraform_build_path}", INDENT1))

        proc: Popen = subprocess.Popen(
            args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=self.terraform_build_path
        )

        # Wait for results
        # Debug always true here because otherwise nothing gets printed and you don't know what the state of the
        # image building is
        output, error = stream_process(proc, True)
        return output, error, proc.returncode

    def gcloud_activate_service_account(self) -> Tuple[str, str, int]:
        args = ["gcloud", "auth", "activate-service-account", "--key-file", self.config.google_cloud.credentials]

        if self.debug:
            print("Executing subprocess:")
            print(indent(f"Command: {subprocess.list2cmdline(args)}", INDENT1))
            print(indent(f"Cwd: {self.api_package_path}", INDENT1))

        proc: Popen = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=self.api_package_path)

        # Wait for results
        # Debug always true here because otherwise nothing gets printed and you don't know what the state of the
        # image building is
        output, error = stream_process(proc, True)
        return output, error, proc.returncode

    def gcloud_builds_submit(self) -> Tuple[str, str, int]:
        # Build the google container image
        project_id = self.config.google_cloud.project_id
        # --gcs-logs-dir is specified to avoid storage.objects.get access error, see:
        # https://github.com/google-github-actions/setup-gcloud/issues/105
        # the _cloudbuild bucket is created already to store the build image
        args = [
            "gcloud",
            "builds",
            "submit",
            "--tag",
            f"gcr.io/{project_id}/observatory-api",
            "--project",
            project_id,
            "--gcs-log-dir",
            f"gs://{project_id}_cloudbuild/logs",
        ]
        if self.debug:
            print("Executing subprocess:")
            print(indent(f"Command: {subprocess.list2cmdline(args)}", INDENT1))
            print(indent(f"Cwd: {self.api_package_path}", INDENT1))

        proc: Popen = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=self.api_package_path)

        # Wait for results
        # Debug always true here because otherwise nothing gets printed and you don't know what the state of the
        # image building is
        output, error = stream_process(proc, True)

        info_filepath = os.path.join(self.terraform_build_path, "api_image_build.txt")
        with open(info_filepath, "w") as f:
            f.writelines(line + "\n" for line in output.splitlines()[-2:])
        return output, error, proc.returncode



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs, James Diprose, Tuan Chien

import contextlib
import copy
import logging
import shutil
from abc import ABC, abstractmethod
from functools import partial
from typing import Any, Callable, Dict, List, Union, Optional

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol

import pendulum
from airflow import DAG
from airflow.exceptions import AirflowException
from airflow.models.baseoperator import chain
from airflow.models.variable import Variable
from airflow.operators.python import PythonOperator, ShortCircuitOperator
from airflow.operators.bash import BaseOperator
from observatory.platform.utils.airflow_utils import (
    AirflowVars,
    check_connections,
    check_variables,
)
from observatory.platform.utils.file_utils import list_files
from observatory.platform.utils.workflow_utils import SubFolder, on_failure_callback, workflow_path
from observatory.platform.utils.release_utils import get_start_end_date, get_datasets, add_dataset_release


class ReleaseFunction(Protocol):
    def __call__(self, release: "AbstractRelease", **kwargs: Any) -> Any:
        ...

    """ 
    :param release: A single instance of an AbstractRelease
    :param kwargs: the context passed from the PythonOperator. See 
    https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed to 
    this argument.
    :return: Any.
    """


class ListReleaseFunction(Protocol):
    def __call__(self, releases: List["AbstractRelease"], **kwargs: Any) -> Any:
        ...

    """ 
    :param releases: A list of AbstractRelease instances
    :param kwargs: the context passed from the PythonOperator. See 
    https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed to 
    this argument.
    :return: Any.
    """


WorkflowFunction = Union[ReleaseFunction, ListReleaseFunction]


class AbstractWorkflow(ABC):
    @abstractmethod
    def add_setup_task(self, func: Callable):
        """Add a setup task, which is used to run tasks before 'Release' objects are created, e.g. checking
        dependencies, fetching available releases etc.

        A setup task has the following properties:
        - Has the signature 'def func(self, **kwargs) -> bool', where
        kwargs is the context passed from the PythonOperator. See https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        - Run by a ShortCircuitOperator, meaning that a setup task can stop a DAG prematurely, e.g. if there is
        nothing to process.
        - func Needs to return a boolean

        :param func: the function that will be called by the ShortCircuitOperator task.
        :return: None.
        """
        pass

    @abstractmethod
    def add_setup_task_chain(self, funcs: List[Callable]):
        """Add a list of setup tasks, which are used to run tasks before 'Release' objects are created, e.g. checking
        dependencies, fetching available releases etc. (See add_setup_task for more info.)

        :param funcs: The list of functions that will be called by the ShortCircuitOperator task.
        :return: None.
        """
        pass

    @abstractmethod
    def add_operator(self, operator: BaseOperator):
        """Add an Apache Airflow operator.

        :param operator: the Apache Airflow operator.
        """
        pass

    @abstractmethod
    def add_operator_chain(self, operators: List[BaseOperator]):
        """Add a list of Apache Airflow operators.

        :param operators: the list of Apache Airflow operator.
        """
        pass

    @abstractmethod
    def add_task(self, func: Callable):
        """Add a task, which is used to process releases. A task has the following properties:

        - Has one of the following signatures 'def func(self, release: Release, **kwargs)' or 'def func(self, releases: List[Release], **kwargs)'
        - kwargs is the context passed from the PythonOperator. See https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed to this argument.
        - Run by a PythonOperator.

        :param func: the function that will be called by the PythonOperator task.
        :return: None.
        """
        pass

    @abstractmethod
    def add_task_chain(self, funcs: List[Callable]):
        """Add a list of tasks, which are used to process releases. (See add_task for more info.)

        :param funcs: The list of functions that will be called by the PythonOperator task.
        :return: None.
        """
        pass

    @abstractmethod
    def task_callable(self, func: WorkflowFunction, **kwargs) -> Any:
        """Invoke a task callable. Creates a Release instance or Release instances and calls the given task method.

        :param func: the task method.
        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed
        to this argument.
        :return: Any.
        """
        pass

    @abstractmethod
    def make_release(self, **kwargs) -> Union["AbstractRelease", List["AbstractRelease"]]:
        """Make a release instance. The release is passed as an argument to the function (WorkflowFunction) that is
        called in 'task_callable'.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed
        to this argument.
        :return: A release instance or list of release instances
        """
        pass

    @abstractmethod
    def make_dag(self) -> DAG:
        """Make an Airflow DAG for a workflow.

        :return: the DAG object.
        """
        pass


def make_task_id(func: Callable, kwargs: Dict) -> str:
    """Set a task_id from a func or kwargs.

    :param func: the task function.
    :param kwargs: the task kwargs parameter.
    :return: the task id.
    """

    task_id_key = "task_id"

    if task_id_key in kwargs:
        task_id = kwargs["task_id"]
    else:
        task_id = func.__name__

    return task_id


class Workflow(AbstractWorkflow):
    RELEASE_INFO = "releases"

    def __init__(
        self,
        dag_id: str,
        start_date: pendulum.DateTime,
        schedule_interval: str,
        catchup: bool = False,
        queue: str = "default",
        max_retries: int = 3,
        max_active_runs: int = 1,
        airflow_vars: list = None,
        airflow_conns: list = None,
        workflow_id: int = None,
        dataset_type_id: str = None,
        tags: Optional[List[str]] = None,
    ):
        """Construct a Workflow instance.

        :param dag_id: the id of the DAG.
        :param start_date: the start date of the DAG.
        :param schedule_interval: the schedule interval of the DAG.
        :param catchup: whether to catchup the DAG or not.
        :param queue: the Airflow queue name.
        :param max_retries: the number of times to retry each task.
        :param max_active_runs: the maximum number of DAG runs that can be run at once.
        :param airflow_vars: list of airflow variable keys, for each variable it is checked if it exists in airflow
        :param airflow_conns: list of airflow connection keys, for each connection it is checked if it exists in airflow
        :param workflow_id: api workflow id.
        :param tags: Optional Airflow DAG tags to add.
        """

        self.dag_id = dag_id
        self.start_date = start_date
        self.schedule_interval = schedule_interval
        self.catchup = catchup
        self.queue = queue
        self.max_retries = max_retries
        self.max_active_runs = max_active_runs
        self.airflow_vars = airflow_vars
        self.airflow_conns = airflow_conns
        self._parallel_tasks = False
        self.workflow_id = workflow_id
        self.dataset_type_id = dataset_type_id
        self.tags = tags

        self.operators = []
        self.default_args = {
            "owner": "airflow",
            "start_date": self.start_date,
            "on_failure_callback": on_failure_callback,
            "retries": self.max_retries,
        }
        self.description = self.__doc__
        self.dag = DAG(
            dag_id=self.dag_id,
            schedule_interval=self.schedule_interval,
            default_args=self.default_args,
            catchup=self.catchup,
            max_active_runs=self.max_active_runs,
            doc_md=self.__doc__,
            tags=self.tags,
        )

    def add_operator(self, operator: BaseOperator):
        """Add an Apache Airflow operator.

        :param operator: the Apache Airflow operator.
        :return: None.
        """

        # Update operator settings
        operator.start_date = self.start_date
        operator.dag = self.dag
        operator.queue = self.queue
        operator.__dict__.update(self.default_args)

        # Add list of tasks while parallel_tasks is set
        if self._parallel_tasks:
            if len(self.operators) == 0 or not isinstance(self.operators[-1], List):
                self.operators.append([operator])
            else:
                self.operators[-1].append(operator)
        # Add single task to the end of the list
        else:
            self.operators.append(operator)

    def add_operator_chain(self, operators: List[BaseOperator]):
        """Add a list of Apache Airflow operators.

        :param operators: the list of Apache Airflow operator.
        """

        for operator in operators:
            self.add_operator(operator)

    def add_setup_task(self, func: Callable, **kwargs):
        """Add a setup task, which is used to run tasks before 'Release' objects are created, e.g. checking
        dependencies, fetching available releases etc.

        A setup task has the following properties:
        - Has the signature 'def func(self, **kwargs) -> bool', where
        kwargs is the context passed from the PythonOperator. See https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        - Run by a ShortCircuitOperator, meaning that a setup task can stop a DAG prematurely, e.g. if there is
        nothing to process.
        - func Needs to return a boolean

        :param func: the function that will be called by the ShortCircuitOperator task.
        :return: None.
        """

        kwargs_ = copy.copy(kwargs)
        kwargs_["task_id"] = make_task_id(func, kwargs)
        op = ShortCircuitOperator(python_callable=func, **kwargs_)
        self.add_operator(op)

    def add_setup_task_chain(self, funcs: List[Callable], **kwargs):
        """Add a list of setup tasks, which are used to run tasks before 'Release' objects are created, e.g. checking
        dependencies, fetching available releases etc. (See add_setup_task for more info.)

        :param funcs: The list of functions that will be called by the ShortCircuitOperator task.
        :return: None.
        """

        for func in funcs:
            self.add_setup_task(func, **copy.copy(kwargs))

    def add_task(
        self,
        func: Callable,
        **kwargs,
    ):
        """Add a task, which is used to process releases. A task has the following properties:

        - Has one of the following signatures 'def func(self, release: Release, **kwargs)' or 'def func(self,
        releases: List[Release], **kwargs)'
        - kwargs is the context passed from the PythonOperator. See https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        - Run by a PythonOperator.

        :param func: the function that will be called by the PythonOperator task.
        :return: None.
        """

        kwargs_ = copy.copy(kwargs)
        kwargs_["task_id"] = make_task_id(func, kwargs)
        op = PythonOperator(python_callable=partial(self.task_callable, func), **kwargs_)
        self.add_operator(op)

    def add_task_chain(self, funcs: List[Callable], **kwargs):
        """Add a list of tasks, which are used to process releases. (See add_task for more info.)

        :param funcs: The list of functions that will be called by the PythonOperator task.
        :return: None.
        """

        for func in funcs:
            self.add_task(func, **copy.copy(kwargs))

    @contextlib.contextmanager
    def parallel_tasks(self):
        """When called, all tasks added to the workflow within the `with` block will run in parallel.
        add_task and add_task_chain can be used with this function.

        :return: None.
        """

        try:
            self._parallel_tasks = True
            yield
        finally:
            self._parallel_tasks = False

    def task_callable(self, func: WorkflowFunction, **kwargs) -> Any:
        """Invoke a task callable. Creates a Release instance and calls the given task method. The result can be
        pulled as an xcom in Airflow.

        :param func: the task method.
        :param kwargs: the context passed from the PythonOperator.
        See https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed
        to this argument.
        :return: Any.
        """

        release = self.make_release(**kwargs)
        result = func(release, **kwargs)
        return result

    def make_dag(self) -> DAG:
        """Make an Airflow DAG for a workflow.

        :return: the DAG object.
        """

        with self.dag:
            chain(*self.operators)

        return self.dag

    def check_dependencies(self, **kwargs) -> bool:
        """Checks the 'workflow' attributes, airflow variables & connections and possibly additional custom checks.

        :param kwargs: The context passed from the PythonOperator.
        :return: None.
        """
        # check that vars and connections are available
        vars_valid = True
        conns_valid = True
        if self.airflow_vars:
            vars_valid = check_variables(*self.airflow_vars)
        if self.airflow_conns:
            conns_valid = check_connections(*self.airflow_conns)

        if not vars_valid or not conns_valid:
            raise AirflowException("Required variables or connections are missing")

        return True

    def add_new_dataset_releases(self, releases, **kwargs):
        """Task to add a new DatasetRelease record in the API after the workflow is done.

        :param releases: Release(s) passed to workflow tasks.
        :param kwargs: kwargs passed to PythonOperator functions.
        """

        if self.workflow_id is None:
            raise Exception("workflow_id must be set")

        if not isinstance(releases, list):
            releases = [releases]

        for release in releases:
            start_date, end_date = get_start_end_date(release)
            datasets = get_datasets(workflow_id=self.workflow_id)

            if len(datasets) == 0:
                raise AirflowException(
                    f"Workflow (workflow_id={self.workflow_id}) has no associated dataset. Check the api db is correctly populated."
                )

            for dataset in datasets:
                add_dataset_release(
                    start_date=start_date,
                    end_date=end_date,
                    dataset_id=dataset.id,
                )


class AbstractRelease(ABC):
    """The abstract release interface"""

    @property
    @abstractmethod
    def download_bucket(self):
        """The download bucket name.

        :return: the download bucket name.
        """
        pass

    @property
    @abstractmethod
    def transform_bucket(self):
        """The transform bucket name.

        :return: the transform bucket name.
        """
        pass

    @property
    @abstractmethod
    def download_folder(self) -> str:
        """The download folder path for the release, e.g. /path/to/workflows/download/{dag_id}/{release_id}/

        :return: the download folder path.
        """
        pass

    @property
    @abstractmethod
    def extract_folder(self) -> str:
        """The extract folder path for the release, e.g. /path/to/workflows/extract/{dag_id}/{release_id}/

        :return: the extract folder path.
        """
        pass

    @property
    @abstractmethod
    def transform_folder(self) -> str:
        """The transform folder path for the release, e.g. /path/to/workflows/transform/{dag_id}/{release_id}/

        :return: the transform folder path.
        """
        pass

    @property
    @abstractmethod
    def download_files(self) -> List[str]:
        """List all files downloaded as a part of this release.

        :return: list of downloaded files.
        """
        pass

    @property
    @abstractmethod
    def extract_files(self) -> List[str]:
        """List all files extracted as a part of this release.

        :return: list of extracted files.
        """
        pass

    @property
    @abstractmethod
    def transform_files(self) -> List[str]:
        """List all files transformed as a part of this release.

        :return: list of transformed files.
        """
        pass

    @abstractmethod
    def cleanup(self) -> None:
        """Delete all files and folders associated with this release.

        :return: None.
        """

        pass


class Release(AbstractRelease):
    """Used to store info on a given release"""

    def __init__(
        self,
        dag_id: str,
        release_id: str,
        download_files_regex: str = None,
        extract_files_regex: str = None,
        transform_files_regex: str = None,
    ):
        """Construct a Release instance

        :param dag_id: the id of the DAG.
        :param release_id: the id of the release.
        :param download_files_regex: regex pattern that is used to find files in download folder
        :param extract_files_regex: regex pattern that is used to find files in extract folder
        :param transform_files_regex: regex pattern that is used to find files in transform folder
        """
        self.dag_id = dag_id
        self.release_id = release_id
        self.download_files_regex = download_files_regex
        self.extract_files_regex = extract_files_regex
        self.transform_files_regex = transform_files_regex

    @property
    def download_folder(self) -> str:
        """The download folder path for the release, e.g. /path/to/workflows/download/{dag_id}/{release_id}/

        :return: the download folder path.
        """
        return workflow_path(SubFolder.downloaded.value, self.dag_id, self.release_id)

    @property
    def extract_folder(self) -> str:
        """The extract folder path for the release, e.g. /path/to/workflows/extract/{dag_id}/{release_id}/

        :return: the extract folder path.
        """
        return workflow_path(SubFolder.extracted.value, self.dag_id, self.release_id)

    @property
    def transform_folder(self) -> str:
        """The transform folder path for the release, e.g. /path/to/workflows/transform/{dag_id}/{release_id}/

        :return: the transform folder path.
        """
        return workflow_path(SubFolder.transformed.value, self.dag_id, self.release_id)

    @property
    def download_files(self) -> List[str]:
        """List all files downloaded as a part of this release.

        :return: list of downloaded files.
        """
        return list_files(self.download_folder, self.download_files_regex)

    @property
    def extract_files(self) -> List[str]:
        """List all files extracted as a part of this release.

        :return: list of extracted files.
        """
        return list_files(self.extract_folder, self.extract_files_regex)

    @property
    def transform_files(self) -> List[str]:
        """List all files transformed as a part of this release.

        :return: list of transformed files.
        """
        return list_files(self.transform_folder, self.transform_files_regex)

    @property
    def download_bucket(self):
        """The download bucket name.

        :return: the download bucket name.
        """
        return Variable.get(AirflowVars.DOWNLOAD_BUCKET)

    @property
    def transform_bucket(self):
        """The transform bucket name.

        :return: the transform bucket name.
        """
        return Variable.get(AirflowVars.TRANSFORM_BUCKET)

    def cleanup(self) -> None:
        """Delete all files and folders associated with this release.

        :return: None.
        """
        for path in [self.download_folder, self.extract_folder, self.transform_folder]:
            try:
                shutil.rmtree(path)
            except FileNotFoundError as e:
                logging.warning(f"No such file or directory {path}: {e}")



# Copyright 2021 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose

import dataclasses
import os
import shutil
from abc import ABC, abstractmethod
from typing import Dict, List

from observatory.platform.utils.jinja2_utils import render_template


class BuilderInterface(ABC):
    """An interface for building Docker and DockerCompose systems."""

    @abstractmethod
    def add_template(self, *, path: str, **kwargs):
        """Add a Jinja template that will be rendered to the build directory before running Docker Compose commands.
        :param path: the path to the Jinja2 template.
        :param kwargs: the kwargs to use when rendering the template.
        :return: None
        """

        pass

    @abstractmethod
    def add_file(self, *, path: str, output_file_name: str):
        """Add a file that will be copied to the build directory before running the Docker Compose commands.
        :param path: a path to the file.
        :param output_file_name: the output file name.
        :return: None
        """

        pass

    @abstractmethod
    def make_files(self) -> None:
        """Render all Jinja templates and copy all files into the build directory.
        :return: None.
        """


def rendered_file_name(template_file_path: str):
    """Make the rendered file name from the Jinja template name.
    :param template_file_path: the file path to the Jinja2 template.
    :return: the output file name.
    """

    return os.path.basename(template_file_path).replace(".jinja2", "")


@dataclasses.dataclass
class File:
    """A file to copy.
    :param path: the path to the file.
    :param output_file_name: the output file.
    """

    path: str
    output_file_name: str


@dataclasses.dataclass
class Template:
    """A Jinja2 Template to render.
    :param path: the path to the Jinja2 template.
    :param kwargs: the kwargs to use when rendering the template.
    """

    path: str
    kwargs: Dict

    @property
    def output_file_name(self) -> str:
        """Make the output file name.
        :return: the output file name.
        """

        return rendered_file_name(self.path)


class Builder(BuilderInterface):
    def __init__(self, *, build_path: str):
        """BuilderInterface implementation.
        :param build_path: the path where the system will be built.
        """

        self.build_path = build_path
        self.templates: List[Template] = []
        self.files: List[File] = []

    def add_template(self, *, path: str, **kwargs):
        """Add a Jinja template that will be rendered to the build directory before running Docker Compose commands.
        :param path: the path to the Jinja2 template.
        :param kwargs: the kwargs to use when rendering the template.
        :return: None
        """

        self.templates.append(Template(path=path, kwargs=kwargs))

    def add_file(self, *, path: str, output_file_name: str):
        """Add a file that will be copied to the build directory before running the Docker Compose commands.
        :param path: a path to the file.
        :param output_file_name: the output file name.
        :return: None
        """

        self.files.append(File(path=path, output_file_name=output_file_name))

    def render_template(self, template: Template, output_file_path: str):
        """Render a file using a Jinja template and save.
        :param template: the template.
        :param output_file_path: the output path.
        :return: None.
        """

        render = render_template(template.path, **template.kwargs)
        with open(output_file_path, "w") as f:
            f.write(render)

    def make_files(self):
        """Render all Jinja templates and copy all files into the build directory.
        :return: None.
        """

        # Clear Docker directory and make build path
        if os.path.exists(self.build_path):
            shutil.rmtree(self.build_path)
        os.makedirs(self.build_path)

        # Render templates
        for template in self.templates:
            output_path = os.path.join(self.build_path, template.output_file_name)
            self.render_template(template, output_path)

        # Copy files
        for file in self.files:
            output_path = os.path.join(self.build_path, file.output_file_name)
            shutil.copy(file.path, output_path)


# Copyright 2021 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Tuan Chien


import json
import os
import re
import shutil
from glob import glob
from pathlib import Path
from typing import List

import pandas as pd


def schema_to_csv(*, schema: List, output: List, prefix: str = ""):
    """
    Convert a schema file to a pandas dataframe, ready for csv export.

    :param schema: Schema list.
    :param output: Output list of rows.
    :param prefix: Name prefixes for when there are nested records.
    """

    for field in schema:
        fname = field["name"] if "name" in field else None
        ftype = field["type"] if "type" in field else None
        fmode = field["mode"] if "mode" in field else None
        fdesc = field["description"] if "description" in field else None

        name = f"{prefix}{fname}"
        row = {"name": name, "type": ftype, "mode": fmode, "description": fdesc}
        output.append(row)

        if ftype == "RECORD":
            ffield = field["fields"]
            schema_to_csv(schema=ffield, output=output, prefix=f"{prefix}{fname}.")


def generate_csv(*, schema_dir):
    """Convert all observatory schema files in JSON format to CSV for inclusion in Sphinx.
    :param schema_dir: Path to schema directory.
    """

    schema_files = glob(os.path.join(schema_dir, "*.json"))
    dst_dir = "schemas"

    if os.path.exists(dst_dir):
        shutil.rmtree(dst_dir)

    Path(dst_dir).mkdir(exist_ok=True, parents=True)

    for schema_file in schema_files:
        filename = os.path.basename(schema_file)
        filename = filename[:-4] + "csv"  # Remove json and add csv suffix
        with open(schema_file, "r") as f:
            data = f.read()
        schema = json.loads(data)
        rows = list()
        schema_to_csv(schema=schema, output=rows)
        df = pd.DataFrame(rows)
        df.to_csv(os.path.join(dst_dir, filename), index=False)


def generate_latest_files():
    """For each schema, generate a schema_latest.csv to indicate it is the latest dataset schema. Then if people just
    want to refer to the latest schema in the documentation they can refer to this link rather than having to update
    the documentation when there is a schema change.

    Does not handle versioned schemas  (wos/scopus). But maybe we should get rid of versioned schemas when those
    telescopes are ported to the new template framework anyway.
    """

    table_files = glob("schemas/*.csv")
    r = re.compile(r"\d{4}-\d{2}-\d{2}")

    # Build a database of schema files
    table_schemas = {}
    for file in table_files:
        filename = os.path.basename(file)
        date_str = r.search(filename)
        date_str_start = date_str.span()[0]
        table_name = filename[: date_str_start - 1]  # -1 accounts for _
        table_schemas[table_name] = table_schemas.get(table_name, list())
        table_schemas[table_name].append(file)

    # Sort schemas
    for table in table_schemas:
        table_schemas[table].sort()

    # Copy the last schema in list since it's latest, to a table_latest.csv
    for table in table_schemas:
        dst_dir = "schemas"
        dst_filename = f"{table}_latest.csv"
        dst_path = os.path.join(dst_dir, dst_filename)
        src_file = table_schemas[table][-1]
        shutil.copyfile(src_file, dst_path)



# Copyright 2019, 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose, Aniek Roelofs

"""
Utility functions for creating the config files that are used with the local development and/or terraform environment.
"""

import glob
import importlib
import logging
import os
import pathlib
from typing import Union

import pendulum
from natsort import natsorted


def module_file_path(module_path: str, nav_back_steps: int = -1) -> str:
    """Get the file path of a module, given the Python import path to the module.

    :param module_path: the Python import path to the module, e.g. observatory.platform.dags
    :param nav_back_steps: the number of steps on the path to step back.
    :return: the file path to the module.
    """

    module = importlib.import_module(module_path)
    file_path = pathlib.Path(module.__file__).resolve()
    return os.path.normpath(str(pathlib.Path(*file_path.parts[:nav_back_steps]).resolve()))


def observatory_home(*subdirs) -> str:
    """Get the .observatory Observatory Platform home directory or subdirectory on the host machine. The home
    directory and subdirectories will be created if they do not exist.

    The default path: ~/.observatory can be overridden with the environment variable OBSERVATORY_HOME.

    :param: subdirs: an optional list of subdirectories.
    :return: the path.
    """

    obs_home = os.environ.get("OBSERVATORY_HOME", None)
    if obs_home is not None:
        path = os.path.join(os.path.expanduser(obs_home), *subdirs)
    else:
        user_home = str(pathlib.Path.home())
        path = os.path.join(user_home, ".observatory", *subdirs)

    if not os.path.exists(path):
        os.makedirs(path, exist_ok=True)

    return path


def terraform_credentials_path() -> str:
    """Get the path to the terraform credentials file that is created with 'terraform login'.

    :return: the path to the terraform credentials file
    """

    return os.path.join(pathlib.Path.home(), ".terraform.d/credentials.tfrc.json")


def find_schema(
    path: str, table_name: str, release_date: pendulum.DateTime, prefix: str = "", ver: str = None
) -> Union[str, None]:

    """Finds a schema file on a given path, with a particular table name, release date and optional prefix.

    If no version string is sepcified, the most recent schema with a date less than or equal to the release date of the
    dataset is returned. If a version string is specified, the most current (date) schema in that series is returned.
    Use the schema_folder function to find the path to the folder containing the schemas.
    For example (grid schemas):
     - grid_2015-09-22.json
     - grid_2016-04-28.json
     - wos_wok5.4_2016-01-01.json  (versioned schema with version 'wok5.4')

    For GRID releases between 2015-09-22 and 2016-04-28 grid_2015-09-22.json is returned and for GRID releases or after
    2016-04-28 grid_2016-04-28.json is returned (until a new schema with a later date is added).
    Unversioned schemas are named with the following pattern: [prefix]table_name_YYYY-MM-DD.json
    Versioned schemas are named as: [prefix]table_name_ver_YYYY-MM-DD.json
    where the things in [] are optional.
    * prefix: an optional prefix for datasets with multiple tables, for instance the Microsoft Academic Graph (MAG)
    dataset schema file names are prefixed with Mag, e.g. MagAffiliations2020-05-21.json, MagAuthors2020-05-21.json.
    The GRID dataset only has one table, so there is no prefix, e.g. grid2015-09-22.json.
    * table_name: the name of the table.
    * ver: version string.
    * YYYY-MM-DD: schema file names end in the release date that the particular schema should be used from in YYYY-MM-DD
    format. For versioned schemas, this is the date of schema creation.
    * prefix and table_name follow the naming conventions of the dataset, e.g. MAG uses CamelCase for tables and fields
    so CamelCase is used. When there is no preference from the dataset then lower snake case is used.

    :param path: the path to search within.
    :param table_name: the name of the table.
    :param release_date: the release date of the table.
    :param prefix: an optional prefix.
    :param ver: Schema version.
    :return: the path to the schema or None if no schema was found.
    """

    logging.info(
        f"Looking for schema with search parameters: analysis_schema_path={path}, "
        f"prefix={prefix}, table_name={table_name}, release_date={release_date}, "
        f"version={ver}"
    )

    # Make search path for schemas
    if ver:
        search_path = os.path.join(path, f"{prefix}{table_name}_{ver}_*.json")
        valid_length = len(prefix) + len(table_name) + len(ver) + 17  # 17 covers YYYY-MM-DD, _, .json
    else:
        search_path = os.path.join(path, f"{prefix}{table_name}_*.json")
        valid_length = len(prefix) + len(table_name) + 16  # 16 covers YYYY-MM-DD, _, .json

    # Find potential schemas with a glob search and sort them naturally
    schema_paths = glob.glob(search_path)

    # Filter out invalid file names
    filtered_paths = list()
    for path in schema_paths:
        file_name = os.path.basename(path)
        if len(file_name) == valid_length:
            filtered_paths.append(path)

    schema_paths = natsorted(filtered_paths)

    # No schemas were found
    if len(schema_paths) == 0:
        logging.error("No schema found.")
        return None

    # Get schemas with dates <= release date
    suffix_len = 5  # .json
    date_str_len = 10  # YYYY-MM-DD
    selected_paths = []
    for path in schema_paths:
        file_name = os.path.basename(path)
        date_str_start = -(date_str_len + suffix_len)
        date_str_end = -suffix_len
        datestr = file_name[date_str_start:date_str_end]
        schema_date = pendulum.parse(datestr)

        if schema_date <= release_date:
            selected_paths.append(path)
        else:
            break

    # Return the schema with the most recent release date
    if len(selected_paths):
        return selected_paths[-1]

    # No schemas were found
    logging.error("No schema found.")
    return None


def utils_templates_path() -> str:
    """Return the path to the util templates.

    :return: the path.
    """

    return module_file_path("observatory.platform.utils.templates")



# Copyright 2020, 2021 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Tuan Chien


import datetime
import os
from time import sleep
from typing import Dict, List, Union

from airflow.exceptions import AirflowException
from airflow.models import DagModel, DagRun
from airflow.sensors.base import BaseSensorOperator
from airflow.utils.session import provide_session
from airflow.utils.state import State
from sqlalchemy.orm.scoping import scoped_session


class DagRunSensor(BaseSensorOperator):
    """
    A sensor for monitoring dag runs from other DAGs.
    Behaviour:
    * If the DAG ID does not exist, throw an exception.
    * Monitors the time period [execution_date - duration, execution_date] for the DAG's dag runs.
      - If there are no runs, then return a success state.
      - If there are runs, look at the latest run in the time period:
        * If the latest run is successful, return a success state.
        * If the latest run is unsuccessful, return a retry state.
    Like other sensors, you can set the poke_interval and timeout in seconds, to control retry
    behaviour.
    """

    def __init__(
        self,
        *,
        external_dag_id: str,
        duration: datetime.timedelta,
        check_exists: bool = True,
        grace_period: datetime.timedelta = datetime.timedelta(minutes=10),
        **kwargs,
    ):
        """
        :param external_dag_id: The DAG ID to monitor.
        :param duration: Size of the window to look back from the current execution date.
        :param check_exists: Whether to perform check for dag existence.
        :param grace_period: If no dag run can be found, sleep for this short grace period and try again.
        :param kwargs: Pass the rest of the parameters to the ExternalTaskSensor.
        """
        super().__init__(**kwargs)

        self.duration = duration
        self.external_dag_id = external_dag_id
        self.check_exists = check_exists
        self.grace_period = grace_period

    @provide_session
    def poke(self, context: Dict, session: scoped_session = None):
        """Sensor interface for determining status of the sensor.

        :param context: Poke context.
        :param session: Database session.
        :return: Sensor poke response.
        """

        success_state = True
        retry_state = False

        if self.check_exists:
            self.check_dag_exists(session)
        execution_date = context["execution_date"]

        date = self.get_latest_execution_date(session=session, execution_date=execution_date)

        # If no execution_date could be found, sleep the grace period, and try again once.This will sleep the duration
        # of the grace period and try again. This is an alternative to scheduling the workflow at a slightly later time
        # to allow Airflow to record the dagrun in the database.
        # Note that this occupies a slot in execution queue.
        if date is None:
            sleep(self.grace_period.total_seconds())
            date = self.get_latest_execution_date(session=session, execution_date=execution_date)

        if not date:
            return success_state

        task_done = self.is_dagrun_done(date=date, session=session)

        if task_done:
            return success_state

        return retry_state

    def get_latest_execution_date(
        self, *, session: scoped_session, execution_date: datetime.datetime
    ) -> Union[datetime.datetime, None]:
        """Get the most recent execution date for a task in a given DAG.

        :param sesssion: Db session.
        :param execution_date: Current execution date.
        :return: Most recent execution date in the window.
        """

        end_date = execution_date
        start_date = end_date - self.duration

        response = (
            session.query(DagRun)
            .filter(
                DagRun.dag_id == self.external_dag_id,
                DagRun.execution_date.between(start_date, end_date),
            )
            .all()
        )

        dates = [r.execution_date for r in response]
        dates.sort(reverse=True)

        if len(dates) > 0:
            return dates[0]

        return None

    def check_dag_exists(self, session: scoped_session):
        """Check if the DAG exists in the DB.  If they do not, then throw an exception.

        :param session: Database session.
        """

        dag_model = session.query(DagModel).filter(DagModel.dag_id == self.external_dag_id).first()
        if dag_model is None or not os.path.exists(dag_model.fileloc):
            raise AirflowException(f"DAG: {self.external_dag_id} does not exist.")

    def is_dagrun_done(self, *, date: datetime.datetime, session: scoped_session) -> bool:
        """Check if the dagrun has finished executing at a given execution date.
        :param date: execution date.
        :return: True if finished, False otherwise.
        """

        success_states = [State.SUCCESS]
        response = (
            session.query(DagRun)
            .filter(
                DagRun.dag_id == self.external_dag_id,
                DagRun.execution_date.in_([date]),
                DagRun.state.in_(success_states),
            )
            .first()
        )

        if response is not None:
            return True

        return False



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs

import pendulum
from airflow import DAG
from airflow.operators.bash import BashOperator

default_args = {"owner": "airflow", "start_date": pendulum.datetime(2020, 8, 10)}

with DAG(dag_id="dummy_telescope", schedule_interval="@daily", default_args=default_args, catchup=True) as dag:
    task1 = BashOperator(task_id="task1", bash_command="echo 'hello'", queue="remote_queue")

    task2 = BashOperator(task_id="task2", bash_command="echo 'world'", queue="remote_queue")

    task1 >> task2



# Copyright 2019 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose

import hashlib
import json
import os
import time
import urllib.error
import urllib.request
from typing import Any, Dict, List, Tuple, Union
from urllib.parse import ParseResult, urljoin, urlparse

import requests
import tldextract
import xmltodict
from importlib_metadata import metadata
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry


def get_url_domain_suffix(url: str) -> str:
    """Extract a URL composed of the domain name and the suffix of the URL. For example, library.curtin.edu would
    become curtin.edu

    :param url: a URL.
    :return: the domain + . + suffix of the URL.
    """

    result = tldextract.extract(url)
    return f"{result.domain}.{result.suffix}"


def unique_id(string: str) -> str:
    """Generate a unique identifier from a string.

    :param string: a string.
    :return: the unique identifier.
    """
    return hashlib.md5(string.encode("utf-8")).hexdigest()


def is_url_absolute(url: Union[str, ParseResult]) -> bool:
    """Return whether a URL is absolute or relative.

    :param url: the URL to test.
    :return: whether the URL is absolute or relative.
    """

    if not isinstance(url, ParseResult):
        url = urlparse(url)

    return bool(url.netloc)


def strip_query_params(url: str) -> str:
    """Remove the query parameters from an absolute URL.

    :param url: a URL.
    :return: the URL with the query parameters removed.
    """
    assert is_url_absolute(url), f"strip_query_params: url {url} is not absolute"
    return urljoin(url, urlparse(url).path)


def retry_session(
    num_retries: int = 3, backoff_factor: float = 0.5, status_forcelist: Tuple = (500, 502, 503, 50)
) -> requests.Session:
    """Create a session object that is able to retry a given number of times.

    :param num_retries: the number of times to retry.
    :param backoff_factor: the factor to use for determining how long to wait before retrying. See
    urllib3.util.retry.Retry for more details.
    :param status_forcelist: which integer status codes (that would normally not trigger a retry) should
    trigger a retry.
    :return: the session.
    """
    session = requests.Session()
    retry = Retry(
        total=num_retries,
        connect=num_retries,
        read=num_retries,
        redirect=num_retries,
        status_forcelist=status_forcelist,
        backoff_factor=backoff_factor,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session


def get_http_text_response(url: str) -> str:
    """Get the text response from an HTTP GET call.

    :param url: URL to query.
    :return: Text response.
    """

    HTTP_OK = 200
    response = retry_session().get(url)
    if response.status_code != HTTP_OK:
        raise ConnectionError(f"Error requesting {url}. Status: {response.status_code}")

    return response.text


def get_http_response_json(url: str) -> Union[List, Dict]:
    """Get the JSON response from an HTTP API call as a dict.

    :param url: URL to query.
    :return: Dictionary of the response.
    """

    response = get_http_text_response(url)
    response_obj = json.loads(response)
    return response_obj


def get_http_response_xml_to_dict(url: str) -> Dict:
    """Get the XML response from an HTTP API call as a dict.

    :param url: URL to query.
    :return: Dictionary of the response.
    """

    response = get_http_text_response(url)
    response_dict = xmltodict.parse(response)
    return response_dict


def wait_for_url(url: str, timeout: int = 60):
    """Wait for a URL to return a 200 status code.

    :param url: the url to wait for.
    :param timeout: the number of seconds to wait before timing out.
    :return: whether the URL returned a 200 status code or not.
    """

    start = time.time()
    started = False
    while True:
        duration = time.time() - start
        if duration >= timeout:
            break

        try:
            if urllib.request.urlopen(url).getcode() == 200:
                started = True
                break
            time.sleep(0.5)
        except ConnectionResetError:
            pass
        except ConnectionRefusedError:
            pass
        except urllib.error.URLError:
            pass

    return started


def get_user_agent(*, package_name: str) -> str:
    """Return a standardised user agent that can be used by custom web clients to indicate which Python
    package they came from.

    :return: User agent string.
    """

    pkg_info = metadata(package_name)
    version = pkg_info.get("Version")
    url = pkg_info.get("Home-page")
    mailto = pkg_info.get("Author-email")
    ua = f"{package_name} v{version} (+{url}; mailto: {mailto})"

    return ua


def get_observatory_http_header(*, package_name: str) -> Dict:
    """Construct a HTTP header with a custom user agent.

    :param package_name: Package name used to fetch metadata for the User Agent.
    """

    user_agent = get_user_agent(package_name=package_name)
    header = {"User-Agent": user_agent}
    return header


def get_filename_from_url(url: str) -> str:
    """Given a download url with the filename part of the url, extract the filename.

    :param url: URL to parse.
    :return: Filename.
    """

    dst_file = os.path.basename(url)
    tail = dst_file.find("?")  # Remove any parameters
    if tail < 0:
        tail = len(dst_file)
    dst_file = dst_file[:tail]

    return dst_file


def get_filename_from_http_header(url: str) -> str:
    """Get a download filename from the Content-Disposition header.

    :param url: Download url.
    :return: Filename.
    """

    response = requests.head(url)
    cd = response.headers["Content-Disposition"]
    return cd



import logging

from airflow.hooks.base import BaseHook
from observatory.platform.utils.airflow_utils import AirflowConns


def make_observatory_api() -> "ObservatoryApi":  # noqa: F821
    """Make the ObservatoryApi object, configuring it with a host and api_key.

    :return: the ObservatoryApi.
    """

    try:
        from observatory.api.client.api.observatory_api import ObservatoryApi
        from observatory.api.client.api_client import ApiClient
        from observatory.api.client.configuration import Configuration
    except ImportError as e:
        logging.error("Please install the observatory-api Python package to use the make_observatory_api function")
        raise e

    # Get connection
    conn = BaseHook.get_connection(AirflowConns.OBSERVATORY_API)

    # Assert connection has required fields
    assert (
        conn.conn_type != "" and conn.conn_type is not None
    ), f"Airflow Connection {AirflowConns.OBSERVATORY_API} conn_type must not be None"
    assert (
        conn.host != "" and conn.host is not None
    ), f"Airflow Connection {AirflowConns.OBSERVATORY_API} host must not be None"

    # Make host
    host = f'{str(conn.conn_type).replace("_", "-").lower()}://{conn.host}'
    if conn.port:
        host += f":{conn.port}"

    # Only api_key when password present in connection
    api_key = None
    if conn.password != "" and conn.password is not None:
        api_key = {"api_key": conn.password}

    # Return ObservatoryApi
    config = Configuration(host=host, api_key=api_key)
    api_client = ApiClient(config)
    return ObservatoryApi(api_client=api_client)



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Tuan Chien, Aniek Roelofs, James Diprose

""" Utility functions that support (almost) all telescopes/the template """

import calendar
import json
import logging
import os
import pathlib
import re
import traceback
from base64 import b64decode
from datetime import datetime, timedelta
from enum import Enum
from typing import Any, List, Optional, Tuple, Union

import paramiko
import pendulum
import pysftp
import six
from airflow import AirflowException
from airflow.hooks.base import BaseHook
from airflow.models import DagBag, DagRun, Variable, XCom
from airflow.secrets.environment_variables import EnvironmentVariablesBackend
from airflow.utils.db import provide_session
from dateutil.relativedelta import relativedelta
from google.cloud import bigquery
from observatory.platform.observatory_config import Environment
from observatory.platform.utils.airflow_utils import (
    AirflowConns,
    AirflowVars,
    send_slack_msg,
)
from observatory.platform.utils.api import make_observatory_api
from observatory.platform.utils.config_utils import find_schema, utils_templates_path
from observatory.platform.utils.gc_utils import (
    bigquery_sharded_table_id,
    copy_bigquery_table,
    create_bigquery_dataset,
    load_bigquery_table,
    run_bigquery_query,
    select_table_shard_dates,
    upload_files_to_cloud_storage,
)
from observatory.platform.utils.jinja2_utils import (
    make_sql_jinja2_filename,
    render_template,
)
from sqlalchemy import and_
from sqlalchemy.orm import Session

ScheduleInterval = Union[str, timedelta, relativedelta]


def fetch_dags_modules() -> dict:
    """Get the dags modules from the Airflow Variable

    :return: Dags modules
    """

    # Try to get value from env variable first, saving costs from GC secret usage
    dags_modules_str = EnvironmentVariablesBackend().get_variable(AirflowVars.DAGS_MODULE_NAMES)
    if not dags_modules_str:
        dags_modules_str = Variable.get(AirflowVars.DAGS_MODULE_NAMES)
    logging.info(f"dags_modules str: {dags_modules_str}")
    dags_modules_ = json.loads(dags_modules_str)
    logging.info(f"dags_modules: {dags_modules_}")
    return dags_modules_


def fetch_dag_bag(path: str, include_examples: bool = False) -> DagBag:
    """Load a DAG Bag from a given path.

    :param path: the path to the DAG bag.
    :param include_examples: whether to include example DAGs or not.
    :return: None.
    """
    logging.info(f"Loading DAG bag from path: {path}")
    dag_bag = DagBag(path, include_examples=include_examples)

    if dag_bag is None:
        raise Exception(f"DagBag could not be loaded from path: {path}")

    if len(dag_bag.import_errors):
        # Collate loading errors as single string and raise it as exception
        results = []
        for path, exception in dag_bag.import_errors.items():
            results.append(f"DAG import exception: {path}\n{exception}\n\n")
        raise Exception("\n".join(results))

    return dag_bag


def workflow_path(*sub_dirs) -> str:
    """Return a path for saving telescope data. Create it if it doesn't exist.
    :param sub_dirs: the subdirectories.
    :return: the path.
    """

    logging.info("workflow_path: requesting data_path variable")
    # Try to get value from env variable first, saving costs from GC secret usage
    data_path = EnvironmentVariablesBackend().get_variable(AirflowVars.DATA_PATH)
    if data_path is None:
        data_path = Variable.get(AirflowVars.DATA_PATH)

    sub_dirs_ = [subdir.value if isinstance(subdir, Enum) else subdir for subdir in sub_dirs]

    # Create telescope path
    path = os.path.join(data_path, "telescopes", *sub_dirs_)
    if not os.path.exists(path):
        os.makedirs(path, exist_ok=True)
    return path


def blob_name(path: str) -> str:
    """Convert a file path into the full path of the Blob, excluding the bucket name.
    E.g.: '/workdir/data/telescopes/transform/dag_id/dag_id_2021_03_01/file.txt' ->
    'telescopes/dag_id/dag_id_2021_03_01/file.txt'

    :param path: the path to the file.
    :return: the full path of the Blob.
    """

    # Remove everything before 'telescopes' and make sure that we are using posix file paths
    # Which use / forward slashes rather than backslashes
    sub_paths = pathlib.Path(os.path.relpath(path, workflow_path())).parts[1:]
    relative_path = pathlib.Path(*sub_paths).as_posix()

    return pathlib.Path(os.path.join("telescopes", relative_path)).as_posix()


def batch_blob_name(path: str) -> str:
    """

    :param path:
    :return:
    """
    return os.path.join(blob_name(path), "*")


def upload_files_from_list(files_list: List[str], bucket_name: str) -> bool:
    """Upload all files in a list to the google cloud download bucket.

    :param files_list: List of full path of files that will be uploaded
    :param bucket_name: The name of the google cloud bucket
    :return: True if upload was successful, else False.
    """
    blob_names = []
    for file_path in files_list:
        blob_names.append(blob_name(file_path))

    success = upload_files_to_cloud_storage(bucket_name, blob_names, files_list)
    if not success:
        raise AirflowException()

    return success


def table_ids_from_path(transform_path: str) -> Tuple[str, str]:
    """
    To create the table ids the basename of the transform path is taken and file extensions are stripped.
    E.g.: /opt/observatory/data/telescopes/transform/telescope/2020_01_01-2020_02_01/telescope.jsonl.gz -> 'telescope'
    and 'telescope_partitions'
    :param transform_path: The full path of a file in the transform folder
    :return: Main table id and partition table id
    """
    logging.info(f"Creating table ids from path: {transform_path}")

    main_table_id = os.path.splitext(pathlib.Path(transform_path).stem)[0]
    partition_table_id = f"{main_table_id}_partitions"

    logging.info(f"Table id: {main_table_id}, partition table id: {partition_table_id}")
    return main_table_id, partition_table_id


def create_date_table_id(table_id: str, date: datetime, partition_type: bigquery.TimePartitioningType):
    """Create a table id string, which includes the date in the correct format corresponding to the partition type.

    :param table_id: The table id
    :param date: The date used for the partition identifier
    :param partition_type: The partition type
    :return: The updated table id
    """
    time_type = bigquery.TimePartitioningType
    type_map = {time_type.HOUR: "%Y%m%d%H", time_type.DAY: "%Y%m%d", time_type.MONTH: "%Y%m", time_type.YEAR: "%Y"}

    date_format = type_map.get(partition_type)
    if date_format is None:
        raise TypeError("Invalid partition type")

    date_str = date.strftime(date_format)

    return f"{table_id}${date_str}"


def prepare_bq_load(
    schema_folder: str,
    dataset_id: str,
    table_id: str,
    release_date: pendulum.DateTime,
    prefix: str,
    schema_version: str,
    dataset_description: str = "",
) -> Tuple[str, str, str, str]:
    """
    Prepare to load data into BigQuery. This will:
     - create the dataset if it does not exist yet
     - get the path to the schema
     - return values of project id, bucket name, and data location
    :param schema_folder: the path to the SQL schemas.
    :param dataset_id: Dataset id.
    :param table_id: Table id.
    :param release_date: The release date used for schema lookup.
    :param prefix: The prefix for the schema.
    :param schema_version: Schema version.
    :param dataset_description: dataset description.
    :return: The project id, bucket name, data location and schema path
    """

    logging.info("requesting project_id variable")
    project_id = Variable.get(AirflowVars.PROJECT_ID)

    logging.info("requesting transform_bucket variable")
    bucket_name = Variable.get(AirflowVars.TRANSFORM_BUCKET)

    logging.info("requesting data_location variable")
    data_location = Variable.get(AirflowVars.DATA_LOCATION)

    # Create dataset
    create_bigquery_dataset(project_id, dataset_id, data_location, dataset_description)

    # Select schema file based on release date
    schema_file_path = find_schema(schema_folder, table_id, release_date, prefix, schema_version)
    if schema_file_path is None:
        exit(os.EX_CONFIG)
    return project_id, bucket_name, data_location, schema_file_path


def prepare_bq_load_v2(
    schema_folder: str,
    project_id: str,
    dataset_id: str,
    dataset_location: str,
    table_id: str,
    release_date: pendulum.DateTime,
    prefix: str,
    schema_version: str,
    dataset_description: str = "",
) -> str:
    """
    Prepare to load data into BigQuery. This will:
     - create the dataset if it does not exist yet
     - get the path to the schema
     - return values of project id, bucket name, and data location
    :param schema_folder: the path to the SQL schemas.
    :param project_id: project id.
    :param dataset_id: Dataset id.
    :param dataset_location: location of dataset.
    :param table_id: Table id.
    :param release_date: The release date used for schema lookup.
    :param prefix: The prefix for the schema.
    :param schema_version: Schema version.
    :param dataset_description: dataset description.
    :return: The project id, bucket name, data location and schema path
    """

    # Create dataset
    dataset_id = dataset_id
    create_bigquery_dataset(project_id, dataset_id, dataset_location, dataset_description)

    # Select schema file based on release date
    schema_file_path = find_schema(schema_folder, table_id, release_date, prefix, schema_version)
    if schema_file_path is None:
        exit(os.EX_CONFIG)
    return schema_file_path


def bq_load_shard(
    schema_folder: str,
    release_date: pendulum.DateTime,
    transform_blob: str,
    dataset_id: str,
    table_id: str,
    source_format: str,
    prefix: str = "",
    schema_version: str = None,
    dataset_description: str = "",
    **load_bigquery_table_kwargs,
):
    """Load data from a specific file (blob) in the transform bucket to a BigQuery shard.
    :param schema_folder: the path to the SQL schema folder.
    :param release_date: Release date.
    :param transform_blob: Name of the transform blob.
    :param dataset_id: Dataset id.
    :param table_id: Table id.
    :param source_format: the format of the data to load into BigQuery.
    :param prefix: The prefix for the schema.
    :param schema_version: Schema version.
    :param dataset_description: description of the BigQuery dataset.
    :return: None.
    """
    project_id, bucket_name, data_location, schema_file_path = prepare_bq_load(
        schema_folder, dataset_id, table_id, release_date, prefix, schema_version, dataset_description
    )

    # Create table id
    table_id = bigquery_sharded_table_id(table_id, release_date)

    # Load BigQuery table
    uri = f"gs://{bucket_name}/{transform_blob}"
    logging.info(f"URI: {uri}")

    success = load_bigquery_table(
        uri,
        dataset_id,
        data_location,
        table_id,
        schema_file_path,
        source_format,
        project_id=project_id,
        **load_bigquery_table_kwargs,
    )
    if not success:
        raise AirflowException()


def bq_load_shard_v2(
    schema_folder: str,
    project_id: str,
    transform_bucket: str,
    transform_blob: str,
    dataset_id: str,
    dataset_location: str,
    table_id: str,
    release_date: pendulum.Date,
    source_format: str,
    prefix: str = "",
    schema_version: str = None,
    dataset_description: str = "",
    **load_bigquery_table_kwargs,
):
    """Load data from a specific file (blob) in the transform bucket to a BigQuery shard.

    :param schema_folder: the path to the SQL schemas folder.
    :param project_id: project id.
    :param transform_bucket: transform bucket name.
    :param transform_blob: Name of the transform blob.
    :param dataset_id: Dataset id.
    :param dataset_location: location of dataset.
    :param table_id: Table id.
    :param release_date: Release date.
    :param source_format: the format of the data to load into BigQuery.
    :param prefix: The prefix for the schema.
    :param schema_version: Schema version.
    :param dataset_description: description of the BigQuery dataset.
    :return: None.
    """

    schema_file_path = prepare_bq_load_v2(
        schema_folder,
        project_id,
        dataset_id,
        dataset_location,
        table_id,
        release_date,
        prefix,
        schema_version,
        dataset_description,
    )

    # Create table id
    table_id = bigquery_sharded_table_id(table_id, release_date)

    # Load BigQuery table
    uri = f"gs://{transform_bucket}/{transform_blob}"
    logging.info(f"URI: {uri}")

    success = load_bigquery_table(
        uri,
        dataset_id,
        dataset_location,
        table_id,
        schema_file_path,
        source_format,
        project_id=project_id,
        **load_bigquery_table_kwargs,
    )
    if not success:
        raise AirflowException()


def bq_load_ingestion_partition(
    schema_folder: str,
    end_date: pendulum.DateTime,
    transform_blob: str,
    dataset_id: str,
    main_table_id: str,
    partition_table_id: str,
    source_format: str,
    prefix: str = "",
    schema_version: str = None,
    dataset_description: str = "",
    partition_type: bigquery.TimePartitioningType = bigquery.TimePartitioningType.DAY,
    **load_bigquery_table_kwargs,
):
    """Load data from a specific file (blob) in the transform bucket to a partition. Since no partition field is
    given it will automatically partition by ingestion datetime.

    :param schema_folder: the path to the SQL schema folder.
    :param end_date: Release end date, used to find the schema and to create table id.
    :param transform_blob: Name of the transform blob.
    :param dataset_id: Dataset id.
    :param main_table_id: Main table id.
    :param partition_table_id: Partition table id (should include date as data in table is overwritten).
    :param source_format: the format of the data to load into BigQuery.
    :param prefix: The prefix for the schema.
    :param schema_version: Schema version.
    :param dataset_description: The description for the dataset
    :param partition_type: The partitioning type (hour, day, month or year)
    :return: None.
    """
    project_id, bucket_name, data_location, schema_file_path = prepare_bq_load(
        schema_folder, dataset_id, main_table_id, end_date, prefix, schema_version, dataset_description
    )

    uri = f"gs://{bucket_name}/{transform_blob}"

    # Include date in table id, so data in table is not overwritten
    partition_table_id = create_date_table_id(partition_table_id, end_date, partition_type)
    success = load_bigquery_table(
        uri,
        dataset_id,
        data_location,
        partition_table_id,
        schema_file_path,
        source_format,
        partition=True,
        partition_type=partition_type,
        project_id=project_id,
        **load_bigquery_table_kwargs,
    )
    if not success:
        raise AirflowException()


def bq_load_partition(
    schema_folder: str,
    project_id: str,
    transform_bucket: str,
    transform_blob: str,
    dataset_id: str,
    dataset_location: str,
    table_id: str,
    release_date: pendulum.DateTime,
    source_format: str,
    partition_type: bigquery.TimePartitioningType,
    prefix: str = "",
    schema_version: str = None,
    dataset_description: str = "",
    partition_field: str = "release_date",
    **load_bigquery_table_kwargs,
):
    """Load data from a specific file (blob) in the transform bucket to a partition.

    :param schema_folder: the path to the SQL schema path.
    :param project_id: project id.
    :param transform_bucket: transform bucket name.
    :param transform_blob: Name of the transform blob.
    :param dataset_id: Dataset id.
    :param dataset_location: location of dataset.
    :param table_id: Table id.
    :param release_date: Release date.
    :param source_format: the format of the data to load into BigQuery.
    :param partition_type: The partitioning type (hour, day, month or year)
    :param prefix: The prefix for the schema.
    :param schema_version: Schema version.
    :param dataset_description: description of the BigQuery dataset.
    :param partition_field: The name of the partition field in the BigQuery table
    :return: None.
    """

    schema_file_path = prepare_bq_load_v2(
        schema_folder,
        project_id,
        dataset_id,
        dataset_location,
        table_id,
        release_date,
        prefix,
        schema_version,
        dataset_description,
    )

    uri = f"gs://{transform_bucket}/{transform_blob}"

    # Include date in table id, so data in table is not overwritten
    table_id = create_date_table_id(table_id, release_date, partition_type)
    success = load_bigquery_table(
        uri,
        dataset_id,
        dataset_location,
        table_id,
        schema_file_path,
        source_format,
        project_id=project_id,
        partition=True,
        partition_field=partition_field,
        partition_type=partition_type,
        **load_bigquery_table_kwargs,
    )
    if not success:
        raise AirflowException()


def bq_delete_old(
    ingestion_date: pendulum.DateTime,
    dataset_id: str,
    main_table_id: str,
    partition_table_id: str,
    merge_partition_field: str,
    bytes_budget: Optional[int] = None,
    project_id: str = None,
):
    """Will run a BigQuery query that deletes rows from the main table that are matched with rows from
    a specific partition of the partition table.
    The query is created from a template and the given info.

    :param ingestion_date: The ingestion date of the partition that will be used.
    :param dataset_id: Dataset id.
    :param main_table_id: Main table id.
    :param partition_table_id: Partition table id.
    :param merge_partition_field: Merge partition field.
    :param bytes_budget: Maximum bytes allowed to be processed.
    :param project_id: GCP project id.
    :return: None.
    """
    ingestion_date = ingestion_date.strftime("%Y-%m-%d")
    # Get merge variables
    dataset_id = dataset_id
    main_table = main_table_id
    partitioned_table = partition_table_id
    merge_condition_field = merge_partition_field

    template_path = os.path.join(utils_templates_path(), make_sql_jinja2_filename("merge_delete_matched"))
    query = render_template(
        template_path,
        dataset=dataset_id,
        main_table=main_table,
        partitioned_table=partitioned_table,
        merge_condition_field=merge_condition_field,
        ingestion_date=ingestion_date,
        project_id=project_id,
    )
    run_bigquery_query(query, bytes_budget=bytes_budget)


def bq_append_from_partition(
    ingestion_date: pendulum.DateTime,
    dataset_id: str,
    main_table_id: str,
    partition_table_id: str,
):
    """Appends rows to the main table by coping a specific partition from the partition table to the main table.

    :param ingestion_date: The ingestion date of the partition that will be used.
    :param dataset_id: Dataset id.
    :param main_table_id: Main table id.
    :param partition_table_id: Partition table id.
    :return: None.
    """
    logging.info("requesting project_id variable")
    project_id = Variable.get(AirflowVars.PROJECT_ID)

    logging.info("requesting data_location variable")
    data_location = Variable.get(AirflowVars.DATA_LOCATION)

    src_table_id = f"{project_id}.{dataset_id}.{partition_table_id}${ingestion_date.strftime('%Y%m%d')}"
    dst_table_id = f"{project_id}.{dataset_id}.{main_table_id}"
    success = copy_bigquery_table(src_table_id, dst_table_id, data_location, bigquery.WriteDisposition.WRITE_APPEND)
    if not success:
        raise AirflowException("Error copying BigQuery table")


def bq_append_from_file(
    schema_folder: str,
    end_date: pendulum.DateTime,
    transform_blob: str,
    dataset_id: str,
    main_table_id: str,
    source_format: str,
    prefix: str = "",
    schema_version: str = None,
    dataset_description: str = "",
    partition_decorator: str = "",
    **load_bigquery_table_kwargs,
):
    """Appends rows to the main table by loading data from a specific file (blob) in the transform bucket.
    :param schema_folder: the path to the SQL schema folder.
    :param end_date: End date.
    :param transform_blob: Name of the transform blob.
    :param dataset_id: Dataset id.
    :param main_table_id: Main table id.
    :param source_format: the format of the data to load into BigQuery.
    :param prefix: The prefix for the schema.
    :param schema_version: Schema version.
    :param dataset_description: The dataset description.
    :return: None.
    """
    project_id, bucket_name, data_location, schema_file_path = prepare_bq_load(
        schema_folder, dataset_id, main_table_id, end_date, prefix, schema_version, dataset_description
    )

    # Load BigQuery table
    uri = f"gs://{bucket_name}/{transform_blob}"
    logging.info(f"URI: {uri}")

    # Append to table table
    table_id = main_table_id + f"${partition_decorator}" if partition_decorator else main_table_id
    success = load_bigquery_table(
        uri,
        dataset_id,
        data_location,
        table_id,
        schema_file_path,
        source_format,
        write_disposition=bigquery.WriteDisposition.WRITE_APPEND,
        project_id=project_id,
        **load_bigquery_table_kwargs,
    )
    if not success:
        raise AirflowException()


def on_failure_callback(context):
    """
    Function that is called on failure of an airflow task. Will create a slack webhook and send a notification.

    :param context: the context passed from the PythonOperator. See
    https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed to
    this  argument.
    :return: None.
    """

    logging.info("requesting environment variable")
    environment = Variable.get(AirflowVars.ENVIRONMENT)

    logging.info("requesting project_id variable")
    project_id = Variable.get(AirflowVars.PROJECT_ID)

    if environment == Environment.develop.value:
        logging.info("Not sending slack notification in develop environment.")
    else:
        exception = context.get("exception")
        if isinstance(exception, Exception):
            formatted_exception = "".join(
                traceback.format_exception(etype=type(exception), value=exception, tb=exception.__traceback__)
            ).strip()
        else:
            formatted_exception = exception

        comments = f"Task failed, exception:\n{formatted_exception}"
        ti = context["ti"]
        execution_date = context["execution_date"]
        send_slack_msg(ti=ti, execution_date=execution_date, comments=comments, project_id=project_id)


class SubFolder(Enum):
    """The type of subfolder to create for telescope data"""

    downloaded = "download"
    extracted = "extract"
    transformed = "transform"


def normalized_schedule_interval(schedule_interval: Optional[str]) -> Optional[ScheduleInterval]:
    """
    Copied from https://github.com/apache/airflow/blob/main/airflow/models/dag.py#L851-L866

    Licensed to the Apache Software Foundation (ASF) under one
    or more contributor license agreements.  See the NOTICE file
    distributed with this work for additional information
    regarding copyright ownership.  The ASF licenses this file
    to you under the Apache License, Version 2.0 (the
    "License"); you may not use this file except in compliance
    with the License.  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing,
    software distributed under the License is distributed on an
    "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
    KIND, either express or implied.  See the License for the
    specific language governing permissions and limitations
    under the License.
    Returns Normalized Schedule Interval. This is used internally by the Scheduler to
    schedule DAGs.

    1. Converts Cron Preset to a Cron Expression (e.g ``@monthly`` to ``0 0 1 * *``)
    2. If Schedule Interval is "@once" return "None"
    3. If not (1) or (2) returns schedule_interval
    """
    cron_presets = {
        "@hourly": "0 * * * *",
        "@daily": "0 0 * * *",
        "@weekly": "0 0 * * 0",
        "@monthly": "0 0 1 * *",
        "@quarterly": "0 0 1 */3 *",
        "@yearly": "0 0 1 1 *",
    }
    if isinstance(schedule_interval, six.string_types) and schedule_interval in cron_presets:
        _schedule_interval = cron_presets.get(schedule_interval)  # type: Optional[ScheduleInterval]
    elif schedule_interval == "@once":
        _schedule_interval = None
    else:
        _schedule_interval = schedule_interval
    return _schedule_interval


def make_sftp_connection() -> pysftp.Connection:
    """Create a SFTP connection using credentials from the airflow SFTP_SERVICE connection.

    :return: SFTP connection
    """
    conn = BaseHook.get_connection(AirflowConns.SFTP_SERVICE)
    host = conn.host

    # Add public host key
    public_key = conn.extra_dejson.get("host_key", None)
    if public_key is not None:
        key = paramiko.RSAKey(data=b64decode(public_key))
        cnopts = pysftp.CnOpts()
        cnopts.hostkeys.add(host, "ssh-rsa", key)
    else:
        cnopts = pysftp.CnOpts()
        cnopts.hostkeys = None

    # set up connection
    return pysftp.Connection(host, port=conn.port, username=conn.login, password=conn.password, cnopts=cnopts)


class SftpFolders:
    def __init__(self, dag_id: str, organisation_name: str, sftp_root: str = "/"):
        """Initialise SftpFolders.

        :param dag_id: the dag id (namespace + organisation name)
        :param organisation_name: the organisation name.
        :param sftp_root: optional root to be added to sftp home path
        """
        self.dag_id = dag_id
        self.organisation_name = organisation_name
        self.sftp_root = sftp_root

    @property
    def sftp_home(self) -> str:
        """Make the SFTP home folder for an organisation.

        :return: the path to the folder.
        """

        organisation_id = make_org_id(self.organisation_name)
        dag_id_prefix = self.dag_id[: -len(organisation_id) - 1]
        return os.path.join(self.sftp_root, "telescopes", dag_id_prefix, organisation_id)

    @property
    def upload(self) -> str:
        """The organisation's SFTP upload folder.

        :return: path to folder.
        """
        return os.path.join(self.sftp_home, "upload")

    @property
    def in_progress(self) -> str:
        """The organisation's SFTP in_progress folder.

        :return: path to folder.
        """
        return os.path.join(self.sftp_home, "in_progress")

    @property
    def finished(self) -> str:
        """The organisation's SFTP finished folder.

        :return: path to folder.
        """
        return os.path.join(self.sftp_home, "finished")

    def move_files_to_in_progress(self, upload_files: Union[list, str]):
        """Move files in list from upload to in-progress folder.

        :param upload_files: File name or list of file names that are in the upload folder and will be moved to the
        in_progress folder (can be full path or just file name)
        :return: None.
        """

        print(f"Files are: {upload_files}")

        if isinstance(upload_files, str):
            upload_files = [upload_files]

        with make_sftp_connection() as sftp:
            sftp.makedirs(self.in_progress)
            for file in upload_files:
                file_name = os.path.basename(file)
                upload_file = os.path.join(self.upload, file_name)
                in_progress_file = os.path.join(self.in_progress, file_name)
                sftp.rename(upload_file, in_progress_file)

    def move_files_to_finished(self, in_progress_files: Union[list, str]):
        """Move files in list from in_progress to finished folder.

        :param in_progress_files: File name or list of file names that are in the in_progress folder and will be moved
        to the finished folder (can be full path or just file name)
        :return: None.
        """
        if isinstance(in_progress_files, str):
            in_progress_files = [in_progress_files]

        with make_sftp_connection() as sftp:
            sftp.makedirs(self.finished)
            for file in in_progress_files:
                file_name = os.path.basename(file)
                in_progress_file = os.path.join(self.in_progress, file_name)
                finished_file = os.path.join(self.finished, file_name)
                sftp.rename(in_progress_file, finished_file)


def make_dag_id(namespace: str, organisation_name: str) -> str:
    """Make a DAG id from a namespace and an organisation name.

    :param namespace: the namespace for the DAG id.
    :param organisation_name: the organisation name.
    :return: the DAG id.
    """

    return f'{namespace}_{organisation_name.strip().lower().replace(" ", "_")}'


def make_org_id(organisation_name: str) -> str:
    """Make an organisation id from the organisation name. Converts the organisation name to lower case,
    strips whitespace and replaces internal spaces with underscores.

    :param organisation_name: the organisation name.
    :return: the organisation id.
    """

    return organisation_name.strip().replace(" ", "_").lower()


def args_list(args) -> list:
    return args


def convert(k: str) -> str:
    """Convert a key name.
    BigQuery specification for field names: Fields must contain only letters, numbers, and underscores, start with a
    letter or underscore, and be at most 128 characters long.
    :param k: Key.
    :return: Converted key.
    """
    # Trim special characters at start:
    k = re.sub("^[^A-Za-z0-9]+", "", k)
    # Replace other special characters (except '_') in remaining string:
    k = re.sub(r"\W+", "_", k)
    return k


def change_keys(obj, convert):
    """Recursively goes through the dictionary obj and replaces keys with the convert function.
    :param obj: Dictionary object.
    :param convert: Convert function.
    :return: Updated dictionary object.
    """
    if isinstance(obj, (str, int, float)):
        return obj
    if isinstance(obj, dict):
        new = obj.__class__()
        for k, v in list(obj.items()):
            new[convert(k)] = change_keys(v, convert)
    elif isinstance(obj, (list, set, tuple)):
        new = obj.__class__(change_keys(v, convert) for v in obj)
    else:
        return obj
    return new


def build_schedule(sched_start_date: pendulum.DateTime, sched_end_date: pendulum.DateTime):
    """Useful for API based data sources.

    Create a fetch schedule to specify what date ranges to use for each API call. Will default to once a month
    for now, but in the future if we are minimising API calls, this can be a more complicated scheme.

    :param sched_start_date: the schedule start date.
    :param sched_end_date: the end date of the schedule.
    :return: list of (section_start_date, section_end_date) pairs from start_date to current Airflow DAG start date.
    """

    schedule = []
    sched_start_date = sched_start_date.date()
    sched_end_date = sched_end_date.date()

    for start_date in pendulum.Period(start=sched_start_date, end=sched_end_date).range("months"):
        last_day = calendar.monthrange(start_date.year, start_date.month)[1]
        end_date = pendulum.date(start_date.year, start_date.month, last_day)
        schedule.append(pendulum.Period(start_date, end_date))

    return schedule


def get_as_list(base: dict, target):
    """Helper function that returns the target as a list.

    :param base: dictionary to query.
    :param target: target key.
    :return: base[target] as a list (if it isn't already).
    """

    if target not in base:
        return list()

    if not isinstance(base[target], list):
        return [base[target]]

    return base[target]


def get_as_list_or_none(base: dict, key, sub_key):
    """Helper function that returns a list or None if key is missing.

    :param base: dictionary to query.
    :param key: target key.
    :param sub_key: sub_key to target.
    :return: entry or None.
    """

    if key not in base or base[key]["@count"] == "0":
        return None

    return get_as_list(base[key], sub_key)


def get_entry_or_none(base: dict, target, var_type=None):
    """Helper function that returns an entry or None if key is missing.

    :param base: dictionary to query.
    :param target: target key.
    :param var_type: Type of variable this is supposed to be (for casting).
    :return: entry or None.
    """

    if target not in base:
        return None

    if var_type is not None:
        return var_type(base[target])

    return base[target]


def add_partition_date(
    list_of_dicts: List[dict],
    partition_date: datetime,
    partition_type: bigquery.TimePartitioningType = bigquery.TimePartitioningType.DAY,
    partition_field: str = "release_date",
):
    """Add a partition date key/value pair to each dictionary in the list of dicts.
    Used to load data into a BigQuery partition.

    :param list_of_dicts: List of dictionaries with original data
    :param partition_date: The partition date
    :param partition_type: The partition type
    :param partition_field: The name of the partition field in the BigQuery table
    :return: Updated list of dicts with partition dates
    """
    if partition_type == bigquery.TimePartitioningType.HOUR:
        partition_date = partition_date.isoformat()
    else:
        partition_date = partition_date.strftime("%Y-%m-%d")

    for entry in list_of_dicts:
        entry[partition_field] = partition_date
    return list_of_dicts


def make_release_date(**kwargs) -> pendulum.DateTime:
    """Make a release date"""

    return kwargs["next_execution_date"].subtract(days=1).start_of("day")


def is_first_dag_run(dag_run: DagRun) -> bool:
    """Whether the DAG Run is the first run or not

    :param dag_run: A Dag Run instance
    :return: Whether the DAG run is the first run or not
    """

    return dag_run.get_previous_dagrun() is None


def make_table_name(
    *, project_id: str, dataset_id: str, table_id: str, end_date: Union[pendulum.DateTime, pendulum.Date], sharded: bool
):
    """Make a BQ table name.
    :param project_id: GCP Project ID.
    :param dataset_id: GCP Dataset ID.
    :param table_id: Table name to convert to the suitable BQ table ID.
    :param end_date: Latest date considered.
    :param sharded: whether the table is sharded or not.
    """

    new_table_name = table_id
    if sharded:
        table_date = select_table_shard_dates(
            project_id=project_id,
            dataset_id=dataset_id,
            table_id=table_id,
            end_date=end_date,
        )[0]
        new_table_name = f"{table_id}{table_date.strftime('%Y%m%d')}"

    return new_table_name


def get_chunks(*, input_list: List[Any], chunk_size: int = 8) -> List[Any]:
    """Generator that splits a list into chunks of a fixed size.

    :param input_list: Input list.
    :param chunk_size: Size of chunks.
    :return: The next chunk from the input list.
    """

    n = len(input_list)
    for i in range(0, n, chunk_size):
        yield input_list[i : i + chunk_size]


@provide_session
def delete_old_xcoms(
    session: Session = None,
    dag_id: str = None,
    execution_date: pendulum.DateTime = None,
    retention_days: int = 31,
):
    """Delete XCom messages created by the DAG with the given ID that are as old or older than than
    execution_date - retention_days.  Defaults to 31 days of retention.

    :param session: DB session.
    :param dag_id: DAG ID.
    :param execution_date: DAG execution date.
    :param retention_days: Days of messages to retain.
    """

    cut_off_date = execution_date.subtract(days=retention_days)
    session.query(XCom).filter(
        and_(
            XCom.dag_id == dag_id,
            XCom.execution_date <= cut_off_date,
        )
    ).delete()



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose, Aniek Roelofs

INDENT1 = 2
INDENT2 = 3
INDENT3 = 4
INDENT4 = 5


def indent(string: str, num_spaces: int) -> str:
    """Left indent a string.

    :param string: the string to indent.
    :param num_spaces: the number of spaces to indent the string with.
    :return: the indented string.
    """

    assert num_spaces > 0, "indent: num_spaces must be > 0"
    return string.rjust(len(string) + num_spaces)


def comment(string: str) -> str:
    """Add a Python comment character in front of a string.
    :param string: String to comment out.
    :return: Commented string.
    """

    return f"# {string}"



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose

import logging
import subprocess
from subprocess import Popen
from typing import List, Tuple, Union

from airflow.exceptions import AirflowException


def wait_for_process(proc: Popen) -> Tuple[str, str]:
    """Wait for a process to finish, returning the std output and std error streams as strings.

    :param proc: the process object.
    :return: std output and std error streams as strings.
    """
    output, error = proc.communicate()
    output = output.decode("utf-8")
    error = error.decode("utf-8")
    return output, error


def stream_process(proc: Popen, debug: bool) -> Tuple[str, str]:
    """Print output while a process is running, returning the std output and std error streams as strings.

    :param proc: the process object.
    :param debug: whether debug info should be displayed.
    :return: std output and std error streams as strings.
    """
    output_concat = ""
    error_concat = ""
    while True:
        for line in proc.stdout:
            output = line.decode("utf-8")
            if debug:
                print(output, end="")
            output_concat += output
        for line in proc.stderr:
            error = line.decode("utf-8")
            print(error, end="")
            error_concat += error
        if proc.poll() is not None:
            break
    return output_concat, error_concat
